"""Content block schema definitions."""

from __future__ import annotations

from collections.abc import Sequence
from datetime import datetime
from typing import Annotated, Literal, Self

from pydantic import Field

from acp.schema.base import AnnotatedObject


Audience = Sequence[Literal["assistant", "user"]]


class BaseResourceContents(AnnotatedObject):
    """Text-based resource contents."""

    mime_type: str | None = None
    """MIME type of the resource."""

    uri: str
    """URI of the resource."""


class TextResourceContents(BaseResourceContents):
    """Text-based resource contents."""

    text: str
    """Text content of the resource."""


class BlobResourceContents(BaseResourceContents):
    """Binary resource contents."""

    blob: str
    """Base64-encoded binary content of the resource."""


class Annotations(AnnotatedObject):
    """Optional annotations for the client.

    The client can use annotations to inform how objects are used or displayed.
    """

    audience: Audience | None = None
    """Audience for the annotated resource."""

    last_modified: str | None = None
    """Last modified timestamp of the annotated resource."""

    priority: float | None = None
    """Priority of the annotated resource."""

    @classmethod
    def optionally_create(
        cls,
        audience: Audience | None = None,
        last_modified: datetime | str | None = None,
        priority: float | None = None,
    ) -> Self | None:
        """Create an annotations object if any value is given, otherwise return None."""
        if audience is None and last_modified is None and priority is None:
            return None
        if isinstance(last_modified, datetime):
            last_modified = last_modified.strftime("%Y-%m-%dT%H:%M:%SZ")
        return cls(audience=audience, last_modified=last_modified, priority=priority)


class BaseContentBlock(AnnotatedObject):
    """Base content block."""

    annotations: Annotations | None = None
    """Annotations for the content block."""


ResourceContents = TextResourceContents | BlobResourceContents


class EmbeddedResourceContentBlock[TResourceContents: ResourceContents = ResourceContents](
    BaseContentBlock
):
    """Complete resource contents embedded directly in the message.

    Preferred for including context as it avoids extra round-trips.

    Requires the `embeddedContext` prompt capability when included in prompts.
    """

    type: Literal["resource"] = Field(default="resource", init=False)

    resource: TResourceContents
    """Resource content that can be embedded in a message."""


class TextContentBlock(BaseContentBlock):
    """Text content.

    May be plain text or formatted with Markdown.

    All agents MUST support text content blocks in prompts.
    Clients SHOULD render this text as Markdown.
    """

    type: Literal["text"] = Field(default="text", init=False)

    text: str
    """Text content of the block."""


class ImageContentBlock(BaseContentBlock):
    """Images for visual context or analysis.

    Requires the `image` prompt capability when included in prompts.
    """

    type: Literal["image"] = Field(default="image", init=False)

    data: str
    """Base64-encoded image data."""

    mime_type: str
    """MIME type of the image."""

    uri: str | None = None
    """URI of the image."""


class AudioContentBlock(BaseContentBlock):
    """Audio data for transcription or analysis.

    Requires the `audio` prompt capability when included in prompts.
    """

    type: Literal["audio"] = Field(default="audio", init=False)

    data: str
    """Base64-encoded audio data."""

    mime_type: str
    """MIME type of the audio."""


# Resource links:
# [@index.js](file:///Users/.../projects/reqwest/examples/wasm_github_fetch/index.js)
# [@wasm](file:///Users/.../projects/reqwest/src/wasm)
# [@error](file:///Users/.../projects/reqwest/src/async_impl/client.rs?symbol=Error#L2661:2661)
# [@error.rs (23:27)](file:///Users/.../projects/reqwest/src/error.rs#L23:27)


class ResourceContentBlock(BaseContentBlock):
    """References to resources that the agent can access.

    All agents MUST support resource links in prompts.
    """

    type: Literal["resource_link"] = Field(default="resource_link", init=False)

    description: str | None = None
    """Description of the resource."""

    mime_type: str | None = None
    """MIME type of the resource."""

    name: str
    """Name of the resource."""

    size: int | None = Field(default=None, ge=0)
    """Size of the resource in bytes."""

    title: str | None = None
    """Title of the resource."""

    uri: str
    """URI of the resource."""


ContentBlock = Annotated[
    (
        TextContentBlock
        | ImageContentBlock
        | AudioContentBlock
        | ResourceContentBlock
        | EmbeddedResourceContentBlock
    ),
    Field(discriminator="type"),
]



"""Base command class with node-type filtering support."""

from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Any

from slashed import SlashedCommand


if TYPE_CHECKING:
    from llmling_agent.messaging import MessageNode


class NodeCommand(SlashedCommand, ABC):
    """Base class for commands with node-type filtering.

    Commands inheriting from this can override `supports_node` to restrict
    which node types they're available for.
    """

    name: str = ""  # Placeholder, must be overridden by concrete subclasses

    @classmethod
    def supports_node(cls, node: MessageNode[Any, Any]) -> bool:
        """Check if this command supports the given node type.

        Override in subclasses to restrict to specific node types.
        Default implementation returns True (supports all nodes).

        Args:
            node: The node to check compatibility with
        """
        return True


class AgentCommand(NodeCommand):
    """Base class for commands that only work with Agent nodes."""

    @classmethod
    def supports_node(cls, node: MessageNode[Any, Any]) -> bool:
        """Check if node is an Agent."""
        from llmling_agent import Agent

        return isinstance(node, Agent)



"""FSSpec filesystem toolset helpers."""

from __future__ import annotations

import difflib
import re

from pydantic_ai import ModelRetry

from llmling_agent.log import get_logger


logger = get_logger(__name__)

# MIME types that should be treated as text
TEXT_MIME_PREFIXES = ("text/", "application/json", "application/xml", "application/javascript")


async def apply_structured_edits(original_content: str, edits_response: str) -> str:
    """Apply structured edits from the agent response."""
    # Parse the edits from the response
    edits_match = re.search(r"<edits>(.*?)</edits>", edits_response, re.DOTALL)
    if not edits_match:
        logger.warning("No edits block found in response")
        return original_content

    edits_content = edits_match.group(1)

    # Find all old_text/new_text pairs
    old_text_pattern = r"<old_text[^>]*>(.*?)</old_text>"
    new_text_pattern = r"<new_text>(.*?)</new_text>"

    old_texts = re.findall(old_text_pattern, edits_content, re.DOTALL)
    new_texts = re.findall(new_text_pattern, edits_content, re.DOTALL)

    if len(old_texts) != len(new_texts):
        logger.warning("Mismatch between old_text and new_text blocks")
        return original_content

    # Apply edits sequentially
    content = original_content
    applied_edits = 0

    failed_matches = []
    multiple_matches = []

    for old_text, new_text in zip(old_texts, new_texts, strict=False):
        old_cleaned = old_text.strip()
        new_cleaned = new_text.strip()

        # Check for multiple matches (ambiguity)
        match_count = content.count(old_cleaned)
        if match_count > 1:
            multiple_matches.append(old_cleaned[:50])
        elif match_count == 1:
            content = content.replace(old_cleaned, new_cleaned, 1)
            applied_edits += 1
        else:
            failed_matches.append(old_cleaned[:50])

    # Raise ModelRetry for specific failure cases
    if applied_edits == 0 and len(old_cleaned) > 0:
        msg = (
            "Some edits were produced but none of them could be applied. "
            "Read the relevant sections of the file again so that "
            "I can perform the requested edits."
        )
        raise ModelRetry(msg)

    if multiple_matches:
        matches_str = ", ".join(multiple_matches)
        msg = (
            f"<old_text> matches multiple positions in the file: {matches_str}... "
            "Read the relevant sections of the file again and extend <old_text> "
            "to be more specific."
        )
        raise ModelRetry(msg)

    logger.info("Applied structured edits", num=applied_edits, total=len(old_texts))
    return content


def get_changed_lines(original_content: str, new_content: str, path: str) -> list[str]:
    old = original_content.splitlines(keepends=True)
    new = new_content.splitlines(keepends=True)
    diff = list(difflib.unified_diff(old, new, fromfile=path, tofile=path, lineterm=""))
    return [line for line in diff if line.startswith(("+", "-"))]


def get_changed_line_numbers(original_content: str, new_content: str) -> list[int]:
    """Extract line numbers where changes occurred for ACP UI highlighting.

    Similar to Claude Code's line tracking for precise change location reporting.
    Returns line numbers in the new content where changes happened.

    Args:
        original_content: Original file content
        new_content: Modified file content

    Returns:
        List of line numbers (1-based) where changes occurred in new content
    """
    old_lines = original_content.splitlines(keepends=True)
    new_lines = new_content.splitlines(keepends=True)
    # Use SequenceMatcher to find changed blocks
    matcher = difflib.SequenceMatcher(None, old_lines, new_lines)
    changed_line_numbers = set()
    for tag, _i1, _i2, j1, j2 in matcher.get_opcodes():
        if tag in ("replace", "insert", "delete"):
            # For replacements and insertions, mark lines in new content
            # For deletions, mark the position where deletion occurred
            if tag == "delete":
                # Mark the line where deletion occurred (or next line if at end)
                line_num = min(j1 + 1, len(new_lines))
                if line_num > 0:
                    changed_line_numbers.add(line_num)
            else:
                # Mark all affected lines in new content
                for line_num in range(j1 + 1, j2 + 1):  # Convert to 1-based
                    changed_line_numbers.add(line_num)

    return sorted(changed_line_numbers)


def is_text_mime(mime_type: str | None) -> bool:
    """Check if a MIME type represents text content."""
    if mime_type is None:
        return False
    return any(mime_type.startswith(prefix) for prefix in TEXT_MIME_PREFIXES)



"""Convert a webpage to markdown using urltomarkdown.herokuapp.com."""

from __future__ import annotations

import urllib.parse
import uuid

import httpx
from pydantic_ai import UserPromptPart
from slashed import CommandContext  # noqa: TC002

from llmling_agent.log import get_logger
from llmling_agent.messaging.context import NodeContext  # noqa: TC001
from llmling_agent_commands.base import NodeCommand
from llmling_agent_server.acp_server.session import ACPSession  # noqa: TC001


logger = get_logger(__name__)


class UrlToMarkdownCommand(NodeCommand):
    """Convert a webpage to markdown using urltomarkdown.herokuapp.com.

    Fetches a web page and converts it to markdown format,
    making it ideal for staging as AI context.
    """

    name = "url-to-markdown"
    category = "docs"

    async def execute_command(
        self,
        ctx: CommandContext[NodeContext[ACPSession]],
        url: str,
        *,
        title: bool = True,
        links: bool = True,
        clean: bool = True,
    ) -> None:
        """Convert a webpage to markdown.

        Args:
            ctx: Command context with ACP session
            url: URL to convert to markdown
            title: Include page title as H1 header
            links: Include links in markdown output
            clean: Clean/filter content before conversion
        """
        session = ctx.context.data
        assert session

        # Generate tool call ID
        tool_call_id = f"url-to-markdown-{uuid.uuid4().hex[:8]}"

        try:
            # Build API URL and parameters
            api_url = "https://urltomarkdown.herokuapp.com/"
            params = {"url": url}

            if title:
                params["title"] = "true"
            if not links:
                params["links"] = "false"
            if not clean:
                params["clean"] = "false"

            # Start tool call
            await session.notifications.tool_call_start(
                tool_call_id=tool_call_id,
                title=f"Converting to markdown: {url}",
                kind="fetch",
            )

            # Make async HTTP request
            async with httpx.AsyncClient() as client:
                response = await client.get(
                    api_url,
                    params=params,
                    timeout=30.0,
                )
                response.raise_for_status()
                markdown_content = response.text

            # Get title from header if available
            page_title = ""
            if "X-Title" in response.headers:
                page_title = urllib.parse.unquote(response.headers["X-Title"])
                page_title = f" - {page_title}"

            # Stage the markdown content for use in agent context
            staged_part = UserPromptPart(
                content=f"Webpage content from {url}{page_title}:\n\n{markdown_content}"
            )
            session.add_staged_parts([staged_part])

            # Send successful result - wrap in code block for proper display
            staged_count = session.get_staged_parts_count()
            await session.notifications.tool_call_progress(
                tool_call_id=tool_call_id,
                status="completed",
                title=f"Webpage converted and staged ({staged_count} total parts)",
                content=[f"```markdown\n{markdown_content}\n```"],
            )

        except httpx.HTTPStatusError as e:
            logger.exception("HTTP error converting URL", url=url, status=e.response.status_code)
            await session.notifications.tool_call_progress(
                tool_call_id=tool_call_id,
                status="failed",
                title=f"HTTP {e.response.status_code}: Failed to convert {url}",
            )
        except httpx.RequestError as e:
            logger.exception("Request error converting URL", url=url)
            await session.notifications.tool_call_progress(
                tool_call_id=tool_call_id,
                status="failed",
                title=f"Network error: {e}",
            )
        except Exception as e:
            logger.exception("Unexpected error converting URL", url=url)
            await session.notifications.tool_call_progress(
                tool_call_id=tool_call_id,
                status="failed",
                title=f"Error: {e}",
            )



"""Knowledge configuration."""

from __future__ import annotations

from pydantic import ConfigDict, Field
from schemez import Schema

from llmling_agent.prompts.prompts import PromptType


class Knowledge(Schema):
    """Collection of context sources for an agent.

    Supports both simple paths and rich resource types for content loading,
    plus LLMling's prompt system for dynamic content generation.

    Docs: https://phil65.github.io/llmling-agent/YAML%20Configuration/knowledge_configuration/
    """

    paths: list[str] = Field(
        default_factory=list,
        examples=[["docs/readme.md", "https://api.example.com/docs"], ["/data/context.txt"]],
        title="Resource paths",
    )
    """Quick access to files and URLs."""

    prompts: list[PromptType] = Field(default_factory=list, title="Dynamic prompts")
    """Prompts for dynamic content generation:
    - StaticPrompt: Fixed message templates
    - DynamicPrompt: Python function-based
    - FilePrompt: File-based with template support
    """

    convert_to_markdown: bool = Field(default=False, title="Convert to markdown")
    """Whether to convert content to markdown when possible."""

    model_config = ConfigDict(frozen=True)

    def get_resources(self) -> list[PromptType | str]:
        """Get all resources."""
        return self.prompts + self.paths



"""Utilities package."""

from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    import jinja2


def setup_env(env: jinja2.Environment) -> None:
    """Used as extension point for the jinjarope environment.

    Args:
        env: The jinjarope environment to extend
    """
    from llmling_agent.agent.agent import Agent
    from llmling_agent.functional import (
        run_agent,
        run_agent_sync,
        get_structured,
        get_structured_multiple,
        pick_one,
    )

    env.globals |= dict(agent=Agent)
    env.filters |= {
        "run_agent": run_agent,
        "run_agent_sync": run_agent_sync,
        "pick_one": pick_one,
        "get_structured": get_structured,
        "get_structured_multiple": get_structured_multiple,
    }



"""Models for response fields and definitions."""

from __future__ import annotations

from pydantic import Field
from schemez import Schema, SchemaDef


class StructuredResponseConfig(Schema):
    """Base class for response definitions.

    Docs: https://phil65.github.io/llmling-agent/YAML%20Configuration/response_configuration/
    """

    response_schema: SchemaDef = Field(
        examples=[
            {
                "type": "import",
                "import_path": "llmling_agent.models.content:ImageBase64Content",
            },
            {
                "type": "inline",
                "fields": {
                    "name": {"type": "str", "description": "User name"},
                    "age": {"type": "int", "description": "User age"},
                },
            },
        ],
        title="Response schema",
    )
    """A model describing the response schema."""

    description: str | None = Field(
        default=None,
        examples=["User profile data", "Search results", "Code analysis output"],
        title="Response description",
    )
    """A description for this response definition."""

    result_tool_name: str = Field(
        default="final_result",
        examples=["final_result", "submit_answer", "return_data"],
        title="Result tool name",
    )
    """The tool name for the Agent tool to create the structured response."""

    result_tool_description: str | None = Field(
        default=None,
        examples=["Submit the final result", "Return structured data"],
        title="Result tool description",
    )
    """The tool description for the Agent tool to create the structured response."""

    output_retries: int | None = Field(default=None, examples=[1, 3, 5], title="Output retries")
    """Retry override. How often the Agent should try to validate the response."""



"""Models for resource information."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Self


if TYPE_CHECKING:
    from mcp.types import Resource as MCPResource


@dataclass
class ResourceInfo:
    """Information about an available resource.

    This class provides essential information about a resource that can be loaded.
    Use the resource name with load_resource() to access the actual content.
    """

    name: str
    """Name of the resource, use this with load_resource()"""

    uri: str
    """URI identifying the resource location"""

    description: str | None = None
    """Optional description of the resource's content or purpose"""

    @classmethod
    async def from_mcp_resource(cls, resource: MCPResource) -> Self:
        """Create ResourceInfo from MCP resource."""
        return cls(name=resource.name, uri=str(resource.uri), description=resource.description)



"""Client ACP Connection."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any, Self

import logfire


if TYPE_CHECKING:
    from types import TracebackType

    from acp.client.protocol import Client
    from acp.connection import StreamObserver
    from acp.schema import (
        AuthenticateRequest,
        CancelNotification,
        ClientMethod,
        CreateTerminalResponse,
        InitializeRequest,
        KillTerminalCommandResponse,
        LoadSessionRequest,
        NewSessionRequest,
        PromptRequest,
        ReadTextFileResponse,
        ReleaseTerminalResponse,
        RequestPermissionResponse,
        SetSessionModelRequest,
        SetSessionModeRequest,
        TerminalOutputResponse,
        WaitForTerminalExitResponse,
        WriteTextFileResponse,
    )

from acp.agent.protocol import Agent
from acp.connection import Connection
from acp.exceptions import RequestError
from acp.schema import (
    AuthenticateResponse,
    CreateTerminalRequest,
    InitializeResponse,
    KillTerminalCommandRequest,
    LoadSessionResponse,
    NewSessionResponse,
    PromptResponse,
    ReadTextFileRequest,
    ReleaseTerminalRequest,
    RequestPermissionRequest,
    SessionNotification,
    SetSessionModelResponse,
    SetSessionModeResponse,
    TerminalOutputRequest,
    WaitForTerminalExitRequest,
    WriteTextFileRequest,
)


if TYPE_CHECKING:
    import asyncio
    from collections.abc import Callable


class ClientSideConnection(Agent):
    """Client-side connection.

    Use when you implement the Client and need to talk to an Agent.

    Args:
      to_client: factory that receives this connection and returns your Client
      input: asyncio.StreamWriter (local -> peer)
      output: asyncio.StreamReader (peer -> local)
    """

    def __init__(
        self,
        to_client: Callable[[Agent], Client],
        input_stream: asyncio.StreamWriter,
        output_stream: asyncio.StreamReader,
        observers: list[StreamObserver] | None = None,
    ) -> None:
        # Build client first so handler can delegate
        client = to_client(self)
        handler = partial(_handle_client_method, client)
        self._conn = Connection(handler, input_stream, output_stream, observers=observers)

    # agent-bound methods (client -> agent)
    async def initialize(self, params: InitializeRequest) -> InitializeResponse:
        dct = params.model_dump(
            mode="json", by_alias=True, exclude_none=True, exclude_defaults=True
        )
        resp = await self._conn.send_request("initialize", dct)
        return InitializeResponse.model_validate(resp)

    async def new_session(self, params: NewSessionRequest) -> NewSessionResponse:
        # Don't exclude_defaults - MCP server 'type' field is required for discriminated unions
        dct = params.model_dump(mode="json", by_alias=True, exclude_none=True)
        resp = await self._conn.send_request("session/new", dct)
        return NewSessionResponse.model_validate(resp)

    async def load_session(self, params: LoadSessionRequest) -> LoadSessionResponse:
        dct = params.model_dump(
            mode="json", by_alias=True, exclude_none=True, exclude_defaults=True
        )
        resp = await self._conn.send_request("session/load", dct)
        payload = resp if isinstance(resp, dict) else {}
        return LoadSessionResponse.model_validate(payload)

    async def set_session_mode(self, params: SetSessionModeRequest) -> SetSessionModeResponse:
        dct = params.model_dump(
            mode="json", by_alias=True, exclude_none=True, exclude_defaults=True
        )
        resp = await self._conn.send_request("session/set_mode", dct)
        payload = resp if isinstance(resp, dict) else {}
        return SetSessionModeResponse.model_validate(payload)

    async def set_session_model(self, params: SetSessionModelRequest) -> SetSessionModelResponse:
        dct = params.model_dump(
            mode="json", by_alias=True, exclude_none=True, exclude_defaults=True
        )
        resp = await self._conn.send_request("session/set_model", dct)
        payload = resp if isinstance(resp, dict) else {}
        return SetSessionModelResponse.model_validate(payload)

    async def authenticate(self, params: AuthenticateRequest) -> AuthenticateResponse:
        dct = params.model_dump(
            mode="json", by_alias=True, exclude_none=True, exclude_defaults=True
        )
        resp = await self._conn.send_request("authenticate", dct)
        payload = resp if isinstance(resp, dict) else {}
        return AuthenticateResponse.model_validate(payload)

    async def prompt(self, params: PromptRequest) -> PromptResponse:
        # Don't exclude_defaults here - the 'type' field in content blocks has a default
        # value but is required for discriminated unions to work
        dct = params.model_dump(mode="json", by_alias=True, exclude_none=True)
        resp = await self._conn.send_request("session/prompt", dct)
        return PromptResponse.model_validate(resp)

    async def cancel(self, params: CancelNotification) -> None:
        dct = params.model_dump(
            mode="json", by_alias=True, exclude_none=True, exclude_defaults=True
        )
        await self._conn.send_notification("session/cancel", dct)

    async def ext_method(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        return await self._conn.send_request(f"_{method}", params)  # type: ignore[no-any-return]

    async def ext_notification(self, method: str, params: dict[str, Any]) -> None:
        await self._conn.send_notification(f"_{method}", params)

    async def close(self) -> None:
        await self._conn.close()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.close()


@logfire.instrument(r"Handle Client Method {method}")
async def _handle_client_method(  # noqa: PLR0911
    client: Client,
    method: ClientMethod | str,
    params: dict[str, Any] | None,
    is_notification: bool,
) -> (
    WriteTextFileResponse
    | ReadTextFileResponse
    | RequestPermissionResponse
    | SessionNotification
    | CreateTerminalResponse
    | TerminalOutputResponse
    | WaitForTerminalExitResponse
    | ReleaseTerminalResponse
    | KillTerminalCommandResponse
    | dict[str, Any]
    | None
):
    """Handle client method calls."""
    match method:
        case "fs/write_text_file":
            write_file_request = WriteTextFileRequest.model_validate(params)
            return await client.write_text_file(write_file_request)
        case "fs/read_text_file":
            read_file_request = ReadTextFileRequest.model_validate(params)
            return await client.read_text_file(read_file_request)
        case "session/request_permission":
            permission_request = RequestPermissionRequest.model_validate(params)
            return await client.request_permission(permission_request)
        case "session/update":
            notification = SessionNotification.model_validate(params)
            await client.session_update(notification)
            return None
        case "terminal/create":
            create_request = CreateTerminalRequest.model_validate(params)
            return await client.create_terminal(create_request)
        case "terminal/output":
            output_request = TerminalOutputRequest.model_validate(params)
            return await client.terminal_output(output_request)
        case "terminal/release":
            release_request = ReleaseTerminalRequest.model_validate(params)
            return await client.release_terminal(release_request)
        case "terminal/wait_for_exit":
            wait_request = WaitForTerminalExitRequest.model_validate(params)
            return await client.wait_for_terminal_exit(wait_request)
        case "terminal/kill":
            kill_request = KillTerminalCommandRequest.model_validate(params)
            return await client.kill_terminal(kill_request)
        case str() if method.startswith("_") and is_notification:
            await client.ext_notification(method[1:], params or {})
            return None
        case str() if method.startswith("_") and is_notification:
            return await client.ext_method(method[1:], params or {})
        case _:
            raise RequestError.method_not_found(method)



"""Configuration models for text chunking."""

from __future__ import annotations

from typing import Annotated, Literal

from pydantic import ConfigDict, Field, model_validator
from schemez import Schema


class BaseChunkerConfig(Schema):
    """Base configuration for text chunkers."""

    type: str = Field(init=False, title="Chunker type")
    """Type identifier for the chunker."""

    chunk_overlap: int = Field(default=200, ge=0, examples=[100, 200, 500], title="Chunk overlap")
    """Number of characters to overlap between chunks."""

    model_config = ConfigDict(frozen=True)


class LangChainChunkerConfig(BaseChunkerConfig):
    """Configuration for LangChain chunkers."""

    type: Literal["langchain"] = Field(default="langchain", init=False)
    """Langchain chunker configuration."""

    chunker_type: Literal["recursive", "markdown", "character"] = Field(
        default="recursive",
        examples=["recursive", "markdown", "character"],
        title="LangChain chunker type",
    )
    """Which LangChain chunker to use."""

    chunk_size: int = Field(default=1000, ge=1, examples=[500, 1000, 2000], title="Chunk size")
    """Target size of chunks."""


class MarkoChunkerConfig(BaseChunkerConfig):
    """Configuration for marko-based markdown chunker."""

    type: Literal["marko"] = Field(default="marko", init=False)
    """Marko chunker configuration."""

    split_on: Literal["headers", "paragraphs", "blocks"] = Field(
        default="headers",
        examples=["headers", "paragraphs", "blocks"],
        title="Split strategy",
    )
    """How to split the markdown."""

    min_header_level: int = Field(
        default=2,
        ge=1,
        le=6,
        examples=[1, 2, 3],
        title="Minimum header level",
    )
    """Minimum header level to split on (if splitting on headers)."""

    combine_small_sections: bool = Field(default=True, title="Combine small sections")
    """Whether to combine small sections with neighbors."""

    min_section_length: int = Field(
        default=100,
        ge=0,
        examples=[50, 100, 200],
        title="Minimum section length",
    )
    """Minimum length for a section before combining."""

    @model_validator(mode="after")
    def validate_section_length(self) -> MarkoChunkerConfig:
        """Ensure min_section_length is only used with combine_small_sections."""
        if not self.combine_small_sections and self.min_section_length > 0:
            raise ValueError("min_section_length only valid when combine_small_sections=True")
        return self


class LlamaIndexChunkerConfig(BaseChunkerConfig):
    """Configuration for LlamaIndex chunkers."""

    type: Literal["llamaindex"] = Field(default="llamaindex", init=False)
    """LlamaIndex chunker configuration."""

    chunker_type: Literal["sentence", "token", "fixed", "markdown"] = Field(
        default="markdown",
        examples=["sentence", "token", "fixed", "markdown"],
        title="LlamaIndex chunker type",
    )
    """Which LlamaIndex chunker to use."""

    chunk_size: int = Field(default=1000, ge=1, examples=[500, 1000, 2000], title="Chunk size")
    """Target size of chunks."""

    include_metadata: bool = Field(default=True, title="Include metadata")
    """Whether to include document metadata in chunks."""

    include_prev_next_rel: bool = Field(default=False, title="Include relationships")
    """Whether to track relationships between chunks."""


# Union type for all chunker configs
ChunkerConfig = Annotated[
    LangChainChunkerConfig | LlamaIndexChunkerConfig | MarkoChunkerConfig,
    Field(discriminator="type"),
]



"""The main Agent. Can do all sort of crazy things."""

from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable
from contextlib import AsyncExitStack, asynccontextmanager
from dataclasses import dataclass, field, replace
import time
from typing import TYPE_CHECKING, Any, Self, TypedDict, TypeVar, overload
from uuid import uuid4

from anyenv import MultiEventHandler, method_spawner
from llmling_models import function_to_model, infer_model
import logfire
from psygnal import Signal
from pydantic import ValidationError
from pydantic._internal import _typing_extra
from pydantic_ai import (
    Agent as PydanticAgent,
    AgentRunResultEvent,
    BaseToolCallPart,
    FunctionToolCallEvent,
    FunctionToolResultEvent,
    PartStartEvent,
    RunContext,
    ToolReturnPart,
)
from pydantic_ai.models import Model

from llmling_agent.agent.events import (
    RichAgentStreamEvent,
    RunStartedEvent,
    StreamCompleteEvent,
    ToolCallCompleteEvent,
)
from llmling_agent.common_types import IndividualEventHandler
from llmling_agent.log import get_logger
from llmling_agent.messaging import ChatMessage, MessageHistory, MessageNode
from llmling_agent.messaging.processing import prepare_prompts
from llmling_agent.prompts.convert import convert_prompts
from llmling_agent.storage import StorageManager
from llmling_agent.talk.stats import MessageStats
from llmling_agent.tools import Tool, ToolManager
from llmling_agent.tools.exceptions import ToolError
from llmling_agent.utils.inspection import call_with_context, get_argument_key
from llmling_agent.utils.now import get_now
from llmling_agent.utils.result_utils import to_type
from llmling_agent.utils.streams import merge_queue_into_iterator
from llmling_agent.utils.tasks import TaskManager


TResult = TypeVar("TResult")


if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Coroutine, Sequence
    from datetime import datetime
    from types import TracebackType

    from anyenv.code_execution import ExecutionEnvironment
    from pydantic_ai import UsageLimits
    from pydantic_ai.output import OutputSpec
    from toprompt import AnyPromptType
    from upath.types import JoinablePathLike

    from llmling_agent.agent import AgentContext
    from llmling_agent.common_types import (
        AgentName,
        BuiltinEventHandlerType,
        EndStrategy,
        ModelType,
        ProcessorCallback,
        PromptCompatible,
        SessionIdType,
        ToolType,
    )
    from llmling_agent.delegation import AgentPool, Team, TeamRun
    from llmling_agent.models.agents import AgentConfig, ToolMode
    from llmling_agent.prompts.prompts import PromptType
    from llmling_agent.resource_providers import ResourceProvider
    from llmling_agent.ui.base import InputProvider
    from llmling_agent_config.knowledge import Knowledge
    from llmling_agent_config.mcp_server import MCPServerConfig
    from llmling_agent_config.session import MemoryConfig, SessionQuery
    from llmling_agent_config.task import Job


logger = get_logger(__name__)
# OutputDataT = TypeVar('OutputDataT', default=str, covariant=True)
NoneType = type(None)


class AgentKwargs(TypedDict, total=False):
    """Keyword arguments for configuring an Agent instance."""

    description: str | None
    model: ModelType
    system_prompt: str | Sequence[str]
    tools: Sequence[ToolType | Tool] | None
    toolsets: Sequence[ResourceProvider] | None
    mcp_servers: Sequence[str | MCPServerConfig] | None
    skills_paths: Sequence[JoinablePathLike] | None
    retries: int
    output_retries: int | None
    end_strategy: EndStrategy
    # context: AgentContext[Any] | None  # x
    session: SessionIdType | SessionQuery | MemoryConfig | bool | int
    input_provider: InputProvider | None
    debug: bool
    event_handlers: Sequence[IndividualEventHandler | BuiltinEventHandlerType] | None
    env: ExecutionEnvironment | None


class Agent[TDeps = None, OutputDataT = str](MessageNode[TDeps, OutputDataT]):
    """The main agent class.

    Generically typed with: LLMLingAgent[Type of Dependencies, Type of Result]
    """

    @dataclass(frozen=True)
    class AgentReset:
        """Emitted when agent is reset."""

        agent_name: AgentName
        previous_tools: dict[str, bool]
        new_tools: dict[str, bool]
        timestamp: datetime = field(default_factory=get_now)

    run_failed = Signal(str, Exception)
    agent_reset = Signal(AgentReset)

    def __init__(  # noqa: PLR0915
        # we dont use AgentKwargs here so that we can work with explicit ones in the ctor
        self,
        name: str = "llmling-agent",
        *,
        deps_type: type[TDeps] | None = None,
        model: ModelType = None,
        output_type: OutputSpec[OutputDataT] = str,  # type: ignore[assignment]
        # context: AgentContext[TDeps] | None = None,
        session: SessionIdType | SessionQuery | MemoryConfig | bool | int = None,
        system_prompt: AnyPromptType | Sequence[AnyPromptType] = (),
        description: str | None = None,
        display_name: str | None = None,
        tools: Sequence[ToolType | Tool] | None = None,
        toolsets: Sequence[ResourceProvider] | None = None,
        mcp_servers: Sequence[str | MCPServerConfig] | None = None,
        resources: Sequence[PromptType | str] = (),
        skills_paths: Sequence[JoinablePathLike] | None = None,
        retries: int = 1,
        output_retries: int | None = None,
        end_strategy: EndStrategy = "early",
        input_provider: InputProvider | None = None,
        parallel_init: bool = True,
        debug: bool = False,
        event_handlers: Sequence[IndividualEventHandler | BuiltinEventHandlerType] | None = None,
        agent_pool: AgentPool[Any] | None = None,
        tool_mode: ToolMode | None = None,
        knowledge: Knowledge | None = None,
        agent_config: AgentConfig | None = None,
        env: ExecutionEnvironment | None = None,
    ) -> None:
        """Initialize agent.

        Args:
            name: Identifier for the agent (used for logging and lookups)
            deps_type: Type of dependencies to use
            model: The default model to use (defaults to GPT-5)
            output_type: The default output type to use (defaults to str)
            context: Agent context with configuration
            session: Memory configuration.
                - None: Default memory config
                - False: Disable message history (max_messages=0)
                - int: Max tokens for memory
                - str/UUID: Session identifier
                - MemoryConfig: Full memory configuration
                - MemoryProvider: Custom memory provider
                - SessionQuery: Session query

            system_prompt: System prompts for the agent
            description: Description of the Agent ("what it can do")
            display_name: Human-readable display name (falls back to name)
            tools: List of tools to register with the agent
            toolsets: List of toolset resource providers for the agent
            mcp_servers: MCP servers to connect to
            resources: Additional resources to load
            skills_paths: Local directories to search for agent-specific skills
            retries: Default number of retries for failed operations
            output_retries: Max retries for result validation (defaults to retries)
            end_strategy: Strategy for handling tool calls that are requested alongside
                          a final result
            input_provider: Provider for human input (tool confirmation / HumanProviders)
            parallel_init: Whether to initialize resources in parallel
            debug: Whether to enable debug mode
            event_handlers: Sequence of event handlers to register with the agent
            agent_pool: AgentPool instance for managing agent resources
            tool_mode: Tool execution mode (None or "codemode")
            knowledge: Knowledge sources for this agent
            agent_config: Agent configuration
            env: Execution environment for code/command execution and filesystem access
        """
        from llmling_agent.agent import AgentContext
        from llmling_agent.agent.interactions import Interactions
        from llmling_agent.agent.sys_prompts import SystemPrompts
        from llmling_agent.models.agents import AgentConfig
        from llmling_agent.models.manifest import AgentsManifest
        from llmling_agent.prompts.conversion_manager import ConversionManager
        from llmling_agent_config.session import MemoryConfig

        self.task_manager = TaskManager()
        self._infinite = False
        self.deps_type = deps_type
        self._manifest = agent_pool.manifest if agent_pool else AgentsManifest()
        ctx = AgentContext[TDeps](
            node_name=name,
            definition=self._manifest,
            config=agent_config or AgentConfig(name=name),
            input_provider=input_provider,
            pool=agent_pool,
        )
        self._context = ctx
        # TODO: use to_structured with tool_name / description?
        self._output_type = to_type(output_type)
        memory_cfg = (
            session if isinstance(session, MemoryConfig) else MemoryConfig.from_value(session)
        )
        # Initialize progress queue before super().__init__()
        self._event_queue = asyncio.Queue[RichAgentStreamEvent[Any]]()
        mcp_servers = list(mcp_servers) if mcp_servers else []
        if ctx and (cfg := ctx.config) and cfg.mcp_servers:
            mcp_servers.extend(cfg.get_mcp_servers())
        super().__init__(
            name=name,
            description=description,
            display_name=display_name,
            enable_logging=memory_cfg.enable,
            mcp_servers=mcp_servers,
            agent_pool=agent_pool,
            event_configs=agent_config.triggers if agent_config else [],
        )

        # Initialize tool manager
        from llmling_agent.agent.builtin_handlers import resolve_event_handlers

        resolved_handlers = resolve_event_handlers(event_handlers)
        self.event_handler = MultiEventHandler[IndividualEventHandler](resolved_handlers)
        all_tools = list(tools or [])
        effective_tool_mode = tool_mode
        self.tools = ToolManager(all_tools, tool_mode=effective_tool_mode)

        # MCP manager will be initialized in __aenter__ and providers added there
        for toolset_provider in toolsets or []:
            self.tools.add_provider(toolset_provider)
        aggregating_provider = self.mcp.get_aggregating_provider()
        self.tools.add_provider(aggregating_provider)

        # # Add local skills provider if directories specified
        # if skills_paths:
        #     from llmling_agent.resource_providers.skills import SkillsResourceProvider

        #     skills_paths = [
        #         Path(d) if isinstance(d, str) else d for d in skills_paths
        #     ]
        #     local_skills_provider = SkillsResourceProvider(
        #         skills_dirs=skills_paths,
        #         name=f"{name}_local_skills",
        #         owner=name,
        #     )
        #     self.tools.add_provider(local_skills_provider)

        # Initialize conversation manager
        resources = list(resources)
        effective_knowledge = knowledge or (ctx.config.knowledge if ctx else None)
        if effective_knowledge:
            resources.extend(effective_knowledge.get_resources())
        storage = agent_pool.storage if agent_pool else StorageManager(self._manifest.storage)
        self.conversation = MessageHistory(
            storage=storage,
            converter=ConversionManager(config=self._manifest.conversion),
            session_config=memory_cfg,
            resources=resources,
        )
        self._model = model
        self._retries = retries
        self._end_strategy: EndStrategy = end_strategy
        self._output_retries = output_retries
        # init variables
        self._debug = debug
        self.parallel_init = parallel_init
        self._name = name
        self._background_task: asyncio.Task[ChatMessage[Any]] | None = None
        self.talk = Interactions(self)
        from anyenv.code_execution import LocalExecutionEnvironment

        self.env = env or LocalExecutionEnvironment()

        # Set up system prompts
        all_prompts: list[AnyPromptType] = []
        if isinstance(system_prompt, (list, tuple)):
            all_prompts.extend(system_prompt)
        elif system_prompt:
            all_prompts.append(system_prompt)
        self.sys_prompts = SystemPrompts(all_prompts, prompt_manager=ctx.prompt_manager)

    def __repr__(self) -> str:
        desc = f", {self.description!r}" if self.description else ""
        return f"Agent({self.name!r}, model={self._model!r}{desc})"

    async def __prompt__(self) -> str:
        typ = self.__class__.__name__
        model = self.model_name or "default"
        parts = [f"Agent: {self.name}", f"Type: {typ}", f"Model: {model}"]
        if self.description:
            parts.append(f"Description: {self.description}")
        parts.extend([await self.tools.__prompt__(), self.conversation.__prompt__()])
        return "\n".join(parts)

    async def __aenter__(self) -> Self:
        """Enter async context and set up MCP servers."""
        try:
            # Collect all coroutines that need to be run
            coros: list[Coroutine[Any, Any, Any]] = []
            coros.append(super().__aenter__())
            coros.extend(self.conversation.get_initialization_tasks())
            if self.parallel_init and coros:
                await asyncio.gather(*coros)
            else:
                for coro in coros:
                    await coro
        except Exception as e:
            msg = "Failed to initialize agent"
            raise RuntimeError(msg) from e
        else:
            return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit async context."""
        await super().__aexit__(exc_type, exc_val, exc_tb)

    @overload
    def __and__(  # if other doesnt define deps, we take the agents one
        self, other: ProcessorCallback[Any] | Team[TDeps] | Agent[TDeps, Any]
    ) -> Team[TDeps]: ...

    @overload
    def __and__(  # otherwise, we dont know and deps is Any
        self, other: ProcessorCallback[Any] | Team[Any] | Agent[Any, Any]
    ) -> Team[Any]: ...

    def __and__(self, other: MessageNode[Any, Any] | ProcessorCallback[Any]) -> Team[Any]:
        """Create sequential team using & operator.

        Example:
            group = analyzer & planner & executor  # Create group of 3
            group = analyzer & existing_group  # Add to existing group
        """
        from llmling_agent.delegation.team import Team

        match other:
            case Team():
                return Team([self, *other.nodes])
            case Callable():
                agent_2 = Agent.from_callback(other)
                agent_2.context.pool = self.context.pool
                return Team([self, agent_2])
            case MessageNode():
                return Team([self, other])
            case _:
                msg = f"Invalid agent type: {type(other)}"
                raise ValueError(msg)

    @overload
    def __or__(self, other: MessageNode[TDeps, Any]) -> TeamRun[TDeps, Any]: ...

    @overload
    def __or__[TOtherDeps](self, other: MessageNode[TOtherDeps, Any]) -> TeamRun[Any, Any]: ...

    @overload
    def __or__(self, other: ProcessorCallback[Any]) -> TeamRun[Any, Any]: ...

    def __or__(self, other: MessageNode[Any, Any] | ProcessorCallback[Any]) -> TeamRun[Any, Any]:
        # Create new execution with sequential mode (for piping)
        from llmling_agent import TeamRun

        if callable(other):
            other = Agent.from_callback(other)
            other.context.pool = self.context.pool

        return TeamRun([self, other])

    @overload
    @classmethod
    def from_callback(
        cls,
        callback: Callable[..., Awaitable[TResult]],
        *,
        name: str | None = None,
        **kwargs: Any,
    ) -> Agent[None, TResult]: ...

    @overload
    @classmethod
    def from_callback(
        cls,
        callback: Callable[..., TResult],
        *,
        name: str | None = None,
        **kwargs: Any,
    ) -> Agent[None, TResult]: ...

    @classmethod
    def from_callback(
        cls,
        callback: ProcessorCallback[Any],
        *,
        name: str | None = None,
        **kwargs: Any,
    ) -> Agent[None, Any]:
        """Create an agent from a processing callback.

        Args:
            callback: Function to process messages. Can be:
                - sync or async
                - with or without context
                - must return str for pipeline compatibility
            name: Optional name for the agent
            kwargs: Additional arguments for agent
        """
        name = name or callback.__name__ or "processor"
        model = function_to_model(callback)
        return_type = _typing_extra.get_function_type_hints(callback).get("return")
        if (  # If async, unwrap from Awaitable
            return_type
            and hasattr(return_type, "__origin__")
            and return_type.__origin__ is Awaitable
        ):
            return_type = return_type.__args__[0]
        return Agent(model=model, name=name, output_type=return_type or str, **kwargs)  # pyright: ignore[reportReturnType]

    @property
    def name(self) -> str:
        """Get agent name."""
        return self._name or "llmling-agent"

    @name.setter
    def name(self, value: str) -> None:
        """Set agent name."""
        self._name = value

    @property  # type: ignore[override]
    def context(self) -> AgentContext[TDeps]:
        """Get agent context."""
        return self._context

    @context.setter
    def context(self, value: AgentContext[TDeps]) -> None:
        """Set agent context and propagate to provider."""
        self._context = value

    def to_structured[NewOutputDataT](
        self,
        output_type: type[NewOutputDataT],
        *,
        tool_name: str | None = None,
        tool_description: str | None = None,
    ) -> Agent[TDeps, NewOutputDataT]:
        """Convert this agent to a structured agent.

        Args:
            output_type: Type for structured responses. Can be:
                - A Python type (Pydantic model)
            tool_name: Optional override for result tool name
            tool_description: Optional override for result tool description

        Returns:
            Typed Agent
        """
        self.log.debug("Setting result type", output_type=output_type)
        self._output_type = to_type(output_type)
        return self  # type: ignore

    def is_busy(self) -> bool:
        """Check if agent is currently processing tasks."""
        return bool(self.task_manager._pending_tasks or self._background_task)

    @property
    def model_name(self) -> str | None:
        """Get the model name in a consistent format."""
        return self._model.model_name if isinstance(self._model, Model) else self._model

    def to_tool(
        self,
        *,
        name: str | None = None,
        reset_history_on_run: bool = True,
        pass_message_history: bool = False,
        parent: Agent[Any, Any] | None = None,
    ) -> Tool[OutputDataT]:
        """Create a tool from this agent.

        Args:
            name: Optional tool name override
            reset_history_on_run: Clear agent's history before each run
            pass_message_history: Pass parent's message history to agent
            parent: Optional parent agent for history/context sharing
        """

        async def wrapped_tool(prompt: str) -> Any:
            if pass_message_history and not parent:
                msg = "Parent agent required for message history sharing"
                raise ToolError(msg)

            if reset_history_on_run:
                self.conversation.clear()

            history = None
            if pass_message_history and parent:
                history = parent.conversation.get_history()
                old = self.conversation.get_history()
                self.conversation.set_history(history)
            result = await self.run(prompt)
            if history:
                self.conversation.set_history(old)
            return result.data

        # Set the correct return annotation dynamically
        wrapped_tool.__annotations__ = {"prompt": str, "return": self._output_type or Any}

        normalized_name = self.name.replace("_", " ").title()
        docstring = f"Get expert answer from specialized agent: {normalized_name}"
        if self.description:
            docstring = f"{docstring}\n\n{self.description}"
        tool_name = name or f"ask_{self.name}"
        wrapped_tool.__doc__ = docstring
        wrapped_tool.__name__ = tool_name

        return Tool.from_callable(
            wrapped_tool,
            name_override=tool_name,
            description_override=docstring,
            source="agent",
        )

    async def get_agentlet[AgentOutputType](
        self,
        tool_choice: str | list[str] | None,
        model: ModelType,
        output_type: type[AgentOutputType] | None,
        input_provider: InputProvider | None = None,
    ) -> PydanticAgent[TDeps, AgentOutputType]:
        """Create pydantic-ai agent from current state."""
        # Monkey patch pydantic-ai to recognize AgentContext

        from llmling_agent.agent.tool_wrapping import wrap_tool

        tools = await self.tools.get_tools(state="enabled", names=tool_choice)
        final_type = to_type(output_type) if output_type not in [None, str] else self._output_type
        actual_model = model or self._model
        model_ = infer_model(actual_model) if isinstance(actual_model, str) else actual_model
        agent = PydanticAgent(
            name=self.name,
            model=model_,
            instructions=await self.sys_prompts.format_system_prompt(self),
            retries=self._retries,
            end_strategy=self._end_strategy,
            output_retries=self._output_retries,
            deps_type=self.deps_type or NoneType,
            output_type=final_type,
        )

        context_for_tools = (
            self.context
            if input_provider is None
            else replace(self.context, input_provider=input_provider)
        )

        for tool in tools:
            wrapped = wrap_tool(tool, context_for_tools)
            if get_argument_key(wrapped, RunContext):
                logger.info("Registering tool: with context", tool_name=tool.name)
                agent.tool(wrapped)
            else:
                logger.info("Registering tool: no context", tool_name=tool.name)
                agent.tool_plain(wrapped)

        return agent  # type: ignore[return-value]

    @overload
    async def run(
        self,
        *prompts: PromptCompatible | ChatMessage[Any],
        output_type: None = None,
        model: ModelType = None,
        store_history: bool = True,
        tool_choice: str | list[str] | None = None,
        usage_limits: UsageLimits | None = None,
        message_id: str | None = None,
        conversation_id: str | None = None,
        message_history: MessageHistory | None = None,
        deps: TDeps | None = None,
        input_provider: InputProvider | None = None,
        wait_for_connections: bool | None = None,
        instructions: str | None = None,
    ) -> ChatMessage[OutputDataT]: ...

    @overload
    async def run[OutputTypeT](
        self,
        *prompts: PromptCompatible | ChatMessage[Any],
        output_type: type[OutputTypeT],
        model: ModelType = None,
        store_history: bool = True,
        tool_choice: str | list[str] | None = None,
        usage_limits: UsageLimits | None = None,
        message_id: str | None = None,
        conversation_id: str | None = None,
        message_history: MessageHistory | None = None,
        deps: TDeps | None = None,
        input_provider: InputProvider | None = None,
        wait_for_connections: bool | None = None,
        instructions: str | None = None,
    ) -> ChatMessage[OutputTypeT]: ...

    @method_spawner  # type: ignore[misc]
    async def run(
        self,
        *prompts: PromptCompatible | ChatMessage[Any],
        output_type: type[Any] | None = None,
        model: ModelType = None,
        store_history: bool = True,
        tool_choice: str | list[str] | None = None,
        usage_limits: UsageLimits | None = None,
        message_id: str | None = None,
        conversation_id: str | None = None,
        message_history: MessageHistory | None = None,
        deps: TDeps | None = None,
        input_provider: InputProvider | None = None,
        wait_for_connections: bool | None = None,
        instructions: str | None = None,
    ) -> ChatMessage[Any]:
        """Run agent with prompt and get response.

        Args:
            prompts: User query or instruction
            output_type: Optional type for structured responses
            model: Optional model override
            store_history: Whether the message exchange should be added to the
                            context window
            tool_choice: Filter tool choice by name
            usage_limits: Optional usage limits for the model
            message_id: Optional message id for the returned message.
                        Automatically generated if not provided.
            conversation_id: Optional conversation id for the returned message.
            message_history: Optional MessageHistory object to
                             use instead of agent's own conversation
            deps: Optional dependencies for the agent
            input_provider: Optional input provider for the agent
            wait_for_connections: Whether to wait for connected agents to complete
            instructions: Optional instructions to override the agent's system prompt

        Returns:
            Result containing response and run information

        Raises:
            UnexpectedModelBehavior: If the model fails or behaves unexpectedly
        """
        # Collect all events through run_stream
        final_message: ChatMessage[Any] | None = None
        async for event in self.run_stream(
            *prompts,
            output_type=output_type,
            model=model,
            store_history=store_history,
            tool_choice=tool_choice,
            usage_limits=usage_limits,
            message_id=message_id,
            conversation_id=conversation_id,
            message_history=message_history,
            deps=deps,
            input_provider=input_provider,
            wait_for_connections=wait_for_connections,
            instructions=instructions,
        ):
            if isinstance(event, StreamCompleteEvent):
                final_message = event.message

        if final_message is None:
            msg = "No final message received from stream"
            raise RuntimeError(msg)

        return final_message

    @method_spawner
    async def run_stream(
        self,
        *prompt: PromptCompatible,
        output_type: type[OutputDataT] | None = None,
        model: ModelType = None,
        tool_choice: str | list[str] | None = None,
        store_history: bool = True,
        usage_limits: UsageLimits | None = None,
        message_id: str | None = None,
        conversation_id: str | None = None,
        message_history: MessageHistory | None = None,
        input_provider: InputProvider | None = None,
        wait_for_connections: bool | None = None,
        deps: TDeps | None = None,
        instructions: str | None = None,
    ) -> AsyncIterator[RichAgentStreamEvent[OutputDataT]]:
        """Run agent with prompt and get a streaming response.

        Args:
            prompt: User query or instruction
            output_type: Optional type for structured responses
            model: Optional model override
            tool_choice: Filter tool choice by name
            store_history: Whether the message exchange should be added to the
                           context window
            usage_limits: Optional usage limits for the model
            message_id: Optional message id for the returned message.
                        Automatically generated if not provided.
            conversation_id: Optional conversation id for the returned message.
            message_history: Optional MessageHistory to use instead of agent's own
            input_provider: Optional input provider for the agent
            wait_for_connections: Whether to wait for connected agents to complete
            deps: Optional dependencies for the agent
            instructions: Optional instructions to override the agent's system prompt
        Returns:
            An async iterator yielding streaming events with final message embedded.

        Raises:
            UnexpectedModelBehavior: If the model fails or behaves unexpectedly
        """
        conversation = message_history if message_history is not None else self.conversation
        message_id = message_id or str(uuid4())
        run_id = str(uuid4())
        user_msg, prompts, original_message = await prepare_prompts(*prompt)
        self.message_received.emit(user_msg)
        start_time = time.perf_counter()
        history_list = conversation.get_history()
        pending_parts = conversation.get_pending_parts()
        yield RunStartedEvent(thread_id=self.conversation_id, run_id=run_id, agent_name=self.name)
        try:
            agentlet = await self.get_agentlet(tool_choice, model, output_type, input_provider)
            content = await convert_prompts(prompts)
            response_msg: ChatMessage[Any] | None = None
            # Prepend pending context parts and convert to pydantic-ai format
            converted = [
                *pending_parts,
                *(i if isinstance(i, str) else i.to_pydantic_ai() for i in content),
            ]
            stream_events = agentlet.run_stream_events(
                converted,
                deps=deps,  # type: ignore[arg-type]
                message_history=[m for run in history_list for m in run.to_pydantic_ai()],
                usage_limits=usage_limits,
                instructions=instructions,
            )

            # Stream events through merge_queue for progress events
            async with merge_queue_into_iterator(stream_events, self._event_queue) as events:
                # Track tool call starts to combine with results later
                pending_tcs: dict[str, BaseToolCallPart] = {}
                async for event in events:  # Call event handlers for all events
                    for handler in self.event_handler._wrapped_handlers:
                        await handler(None, event)

                    yield event  # type: ignore[misc]
                    match event:
                        case (
                            PartStartEvent(part=BaseToolCallPart() as tool_part)
                            | FunctionToolCallEvent(part=tool_part)
                        ):
                            # Store tool call start info for later combination with result
                            pending_tcs[tool_part.tool_call_id] = tool_part
                        case FunctionToolResultEvent(tool_call_id=call_id) as result_event:
                            # Check if we have a pending tool call to combine with
                            if call_info := pending_tcs.pop(call_id, None):
                                # Create and yield combined event
                                combined_event = ToolCallCompleteEvent(
                                    tool_name=call_info.tool_name,
                                    tool_call_id=call_id,
                                    tool_input=call_info.args_as_dict(),
                                    tool_result=result_event.result.content
                                    if isinstance(result_event.result, ToolReturnPart)
                                    else result_event.result,
                                    agent_name=self.name,
                                    message_id=message_id,
                                )
                                yield combined_event
                        case AgentRunResultEvent():
                            # Capture final result data, Build final response message
                            response_time = time.perf_counter() - start_time
                            response_msg = await ChatMessage.from_run_result(
                                event.result,
                                agent_name=self.name,
                                message_id=message_id,
                                conversation_id=conversation_id or user_msg.conversation_id,
                                response_time=response_time,
                            )

            # Only finalize if we got a result (stream may exit early on error)
            if response_msg is None:
                msg = "Stream completed without producing a result"
                raise RuntimeError(msg)  # noqa: TRY301
            # Apply forwarding logic if needed
            if original_message:
                response_msg = response_msg.forwarded(original_message)
            # Send additional enriched completion event
            yield StreamCompleteEvent(message=response_msg)  # pyright: ignore[reportReturnType]
            self.message_sent.emit(response_msg)
            await self.log_message(response_msg)
            if store_history:
                conversation.add_chat_messages([user_msg, response_msg])
            await self.connections.route_message(response_msg, wait=wait_for_connections)

        except Exception as e:
            self.log.exception("Agent stream failed")
            self.run_failed.emit("Agent stream failed", e)
            raise

    async def run_iter(
        self,
        *prompt_groups: Sequence[PromptCompatible],
        output_type: type[OutputDataT] | None = None,
        model: ModelType = None,
        store_history: bool = True,
        wait_for_connections: bool | None = None,
    ) -> AsyncIterator[ChatMessage[OutputDataT]]:
        """Run agent sequentially on multiple prompt groups.

        Args:
            prompt_groups: Groups of prompts to process sequentially
            output_type: Optional type for structured responses
            model: Optional model override
            store_history: Whether to store in conversation history
            wait_for_connections: Whether to wait for connected agents

        Yields:
            Response messages in sequence

        Example:
            questions = [
                ["What is your name?"],
                ["How old are you?", image1],
                ["Describe this image", image2],
            ]
            async for response in agent.run_iter(*questions):
                print(response.content)
        """
        for prompts in prompt_groups:
            response = await self.run(
                *prompts,
                output_type=output_type,
                model=model,
                store_history=store_history,
                wait_for_connections=wait_for_connections,
            )
            yield response  # pyright: ignore

    @method_spawner
    async def run_job(
        self,
        job: Job[TDeps, str | None],
        *,
        store_history: bool = True,
        include_agent_tools: bool = True,
    ) -> ChatMessage[OutputDataT]:
        """Execute a pre-defined task.

        Args:
            job: Job configuration to execute
            store_history: Whether the message exchange should be added to the
                           context window
            include_agent_tools: Whether to include agent tools
        Returns:
            Job execution result

        Raises:
            JobError: If task execution fails
            ValueError: If task configuration is invalid
        """
        from llmling_agent.tasks import JobError

        if job.required_dependency is not None:  # noqa: SIM102
            if not isinstance(self.context.data, job.required_dependency):
                msg = (
                    f"Agent dependencies ({type(self.context.data)}) "
                    f"don't match job requirement ({job.required_dependency})"
                )
                raise JobError(msg)

        # Load task knowledge
        if job.knowledge:
            # Add knowledge sources to context
            for source in list(job.knowledge.paths):
                await self.conversation.load_context_source(source)
            for prompt in job.knowledge.prompts:
                await self.conversation.load_context_source(prompt)
        try:
            # Register task tools temporarily
            tools = job.get_tools()
            async with self.tools.temporary_tools(tools, exclusive=not include_agent_tools):
                # Execute job with job-specific tools
                return await self.run(await job.get_prompt(), store_history=store_history)

        except Exception as e:
            self.log.exception("Task execution failed", error=str(e))
            msg = f"Task execution failed: {e}"
            raise JobError(msg) from e

    async def run_in_background(
        self,
        *prompt: PromptCompatible,
        max_count: int | None = None,
        interval: float = 1.0,
        **kwargs: Any,
    ) -> asyncio.Task[ChatMessage[OutputDataT] | None]:
        """Run agent continuously in background with prompt or dynamic prompt function.

        Args:
            prompt: Static prompt or function that generates prompts
            max_count: Maximum number of runs (None = infinite)
            interval: Seconds between runs
            **kwargs: Arguments passed to run()
        """
        self._infinite = max_count is None

        async def _continuous() -> ChatMessage[Any]:
            count = 0
            self.log.debug("Starting continuous run", max_count=max_count, interval=interval)
            latest = None
            while max_count is None or count < max_count:
                try:
                    current_prompts = [
                        call_with_context(p, self.context, **kwargs) if callable(p) else p
                        for p in prompt
                    ]
                    self.log.debug("Generated prompt", iteration=count)
                    latest = await self.run(current_prompts, **kwargs)
                    self.log.debug("Run continuous result", iteration=count)

                    count += 1
                    await asyncio.sleep(interval)
                except asyncio.CancelledError:
                    self.log.debug("Continuous run cancelled")
                    break
                except Exception:
                    count += 1
                    self.log.exception("Background run failed")
                    await asyncio.sleep(interval)
            self.log.debug("Continuous run completed", iterations=count)
            return latest  # type: ignore[return-value]

        await self.stop()  # Cancel any existing background task
        task = asyncio.create_task(_continuous(), name=f"background_{self.name}")
        self.log.debug("Started background task", task_name=task.get_name())
        self._background_task = task
        return task

    async def stop(self) -> None:
        """Stop continuous execution if running."""
        if self._background_task and not self._background_task.done():
            self._background_task.cancel()
            await self._background_task
            self._background_task = None

    async def wait(self) -> ChatMessage[OutputDataT]:
        """Wait for background execution to complete."""
        if not self._background_task:
            msg = "No background task running"
            raise RuntimeError(msg)
        if self._infinite:
            msg = "Cannot wait on infinite execution"
            raise RuntimeError(msg)
        try:
            return await self._background_task
        finally:
            self._background_task = None

    async def share(
        self,
        target: Agent[TDeps, Any],
        *,
        tools: list[str] | None = None,
        history: bool | int | None = None,  # bool or number of messages
        token_limit: int | None = None,
    ) -> None:
        """Share capabilities and knowledge with another agent.

        Args:
            target: Agent to share with
            tools: List of tool names to share
            history: Share conversation history:
                    - True: Share full history
                    - int: Number of most recent messages to share
                    - None: Don't share history
            token_limit: Optional max tokens for history

        Raises:
            ValueError: If requested items don't exist
            RuntimeError: If runtime not available for resources
        """
        # Share tools if requested
        for name in tools or []:
            tool = await self.tools.get_tool(name)
            meta = {"shared_from": self.name}
            target.tools.register_tool(tool.callable, metadata=meta)

        # Share history if requested
        if history:
            history_text = await self.conversation.format_history(
                max_tokens=token_limit,
                num_messages=history if isinstance(history, int) else None,
            )
            target.conversation.add_context_message(
                history_text, source=self.name, metadata={"type": "shared_history"}
            )

    def register_worker(
        self,
        worker: MessageNode[Any, Any],
        *,
        name: str | None = None,
        reset_history_on_run: bool = True,
        pass_message_history: bool = False,
    ) -> Tool:
        """Register another agent as a worker tool."""
        return self.tools.register_worker(
            worker,
            name=name,
            reset_history_on_run=reset_history_on_run,
            pass_message_history=pass_message_history,
            parent=self if pass_message_history else None,
        )

    def set_model(self, model: ModelType) -> None:
        """Set the model for this agent.

        Args:
            model: New model to use (name or instance)

        """
        self._model = model

    async def reset(self) -> None:
        """Reset agent state (conversation history and tool states)."""
        old_tools = await self.tools.list_tools()
        self.conversation.clear()
        await self.tools.reset_states()
        new_tools = await self.tools.list_tools()

        event = self.AgentReset(
            agent_name=self.name,
            previous_tools=old_tools,
            new_tools=new_tools,
        )
        self.agent_reset.emit(event)

    async def get_stats(self) -> MessageStats:
        """Get message statistics (async version)."""
        messages = await self.get_message_history()
        return MessageStats(messages=messages)

    @asynccontextmanager
    async def temporary_state[T](
        self,
        *,
        system_prompts: list[AnyPromptType] | None = None,
        output_type: type[T] | None = None,
        replace_prompts: bool = False,
        tools: list[ToolType] | None = None,
        replace_tools: bool = False,
        history: list[AnyPromptType] | SessionQuery | None = None,
        replace_history: bool = False,
        pause_routing: bool = False,
        model: ModelType | None = None,
    ) -> AsyncIterator[Self | Agent[T]]:
        """Temporarily modify agent state.

        Args:
            system_prompts: Temporary system prompts to use
            output_type: Temporary output type to use
            replace_prompts: Whether to replace existing prompts
            tools: Temporary tools to make available
            replace_tools: Whether to replace existing tools
            history: Conversation history (prompts or query)
            replace_history: Whether to replace existing history
            pause_routing: Whether to pause message routing
            model: Temporary model override
        """
        old_model = self._model
        if output_type:
            old_type = self._output_type
            self.to_structured(output_type)
        async with AsyncExitStack() as stack:
            if system_prompts is not None:  # System prompts
                await stack.enter_async_context(
                    self.sys_prompts.temporary_prompt(system_prompts, exclusive=replace_prompts)
                )

            if tools is not None:  # Tools
                await stack.enter_async_context(
                    self.tools.temporary_tools(tools, exclusive=replace_tools)
                )

            if history is not None:  # History
                await stack.enter_async_context(
                    self.conversation.temporary_state(history, replace_history=replace_history)
                )

            if pause_routing:  # Routing
                await stack.enter_async_context(self.connections.paused_routing())

            elif model is not None:  # Model
                self._model = model

            try:
                yield self
            finally:  # Restore model
                if model is not None and old_model:
                    self._model = old_model
                if output_type:
                    self.to_structured(old_type)

    async def validate_against(
        self,
        prompt: str,
        criteria: type[OutputDataT],
        **kwargs: Any,
    ) -> bool:
        """Check if agent's response satisfies stricter criteria."""
        result = await self.run(prompt, **kwargs)
        try:
            criteria.model_validate(result.content.model_dump())  # type: ignore
        except ValidationError:
            return False
        else:
            return True


if __name__ == "__main__":
    import logging

    logfire.configure()
    logfire.instrument_pydantic_ai()
    logging.basicConfig(handlers=[logfire.LogfireLoggingHandler()])
    sys_prompt = "Open browser with google,"
    _model = "openai:gpt-5-nano"

    async def handle_events(ctx: RunContext, event: Any) -> None:
        print(f"[EVENT] {type(event).__name__}: {event}")

    agent = Agent(model=_model, tools=["webbrowser.open"], event_handlers=[handle_events])
    result = agent.run.sync(sys_prompt)  # type: ignore[attr-defined]



"""Code execution provider with secure tool isolation via FastAPI server."""

from __future__ import annotations

import asyncio
import contextlib
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Self

from llmling_agent.log import get_logger


if TYPE_CHECKING:
    from collections.abc import Sequence
    import socket
    from types import TracebackType

    from anyenv.code_execution.base import ExecutionEnvironment
    from anyenv.code_execution.configs import ExecutionEnvironmentConfig
    from anyenv.code_execution.models import ServerInfo
    from fastapi import FastAPI
    from schemez import ToolsetCodeGenerator
    import uvicorn

    from llmling_agent.tools.base import Tool


logger = get_logger(__name__)


@dataclass
class RemoteCodeExecutor:
    """Provides secure code execution with tool access via FastAPI server.

    Architecture:
    - FastAPI server runs in HOST environment with tool routes
    - User code runs in SANDBOX environment (Docker, E2B, etc.)
    - Sandbox makes HTTP calls to server for tool execution
    """

    toolset_generator: ToolsetCodeGenerator
    """Code generator for tools."""

    execution_env: ExecutionEnvironment
    """Execution environment for running code."""

    @classmethod
    def from_tools(
        cls,
        tools: Sequence[Tool],
        env_config: ExecutionEnvironmentConfig,
        server_host: str = "localhost",
        server_port: int = 8000,
        include_docstrings: bool = True,
    ) -> RemoteCodeExecutor:
        """Create provider from tools and environment configuration.

        Args:
            tools: Tools to make available for code execution
            env_config: Execution environment configuration
            server_host: Host for FastAPI server
            server_port: Port for FastAPI server
            include_docstrings: Include function docstrings in documentation

        Returns:
            RemoteCodeExecutor instance
        """
        from llmling_agent.resource_providers.codemode.helpers import tools_to_codegen

        toolset_gen = tools_to_codegen(tools, include_docstrings)
        server_handler = ToolServerLifecycleHandler(toolset_gen, server_host, server_port)
        execution_env = env_config.get_provider(server_handler)
        return cls(toolset_gen, execution_env)

    def get_tool_description(self) -> str:
        """Get comprehensive description of available tools."""
        return self.toolset_generator.generate_tool_description()

    async def execute_code(self, code: str) -> Any:
        """Execute code with tools available via HTTP API.

        Args:
            code: Python code to execute

        Returns:
            Execution result from the environment
        """
        return await self.execution_env.execute(code)

    async def __aenter__(self) -> Self:
        """Async context manager entry."""
        await self.execution_env.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Async context manager exit."""
        return await self.execution_env.__aexit__(exc_type, exc_val, exc_tb)


class ToolServerLifecycleHandler:
    """Manages FastAPI server lifecycle for tool access."""

    def __init__(
        self,
        toolset_generator: ToolsetCodeGenerator,
        host: str = "localhost",
        port: int = 8000,
    ) -> None:
        self.toolset_generator = toolset_generator
        self.host = host
        self.port = port  # Will be set when socket is created
        self.app: FastAPI | None = None
        self.server: uvicorn.Server | None = None
        self._server_task: asyncio.Task[None] | None = None
        self._socket: socket.socket | None = None

    async def __aenter__(self) -> ServerInfo:
        """Start FastAPI server with tool routes."""
        from anyenv.code_execution.models import ServerInfo
        from fastapi import FastAPI

        from llmling_agent.utils.network import _create_socket

        if self.server is not None:
            return ServerInfo(url=f"http://{self.host}:{self.port}", port=self.port)
        # Create socket and get actual port
        self._socket, self.port = _create_socket(self.port)
        # Create FastAPI app
        self.app = FastAPI(title="Tool Server", description="Generated tool endpoints")

        # Add tool routes
        self.toolset_generator.add_all_routes(self.app, "/tools")

        import uvicorn

        config = uvicorn.Config(
            self.app,
            log_level="error",  # Reduce log noise
            access_log=False,
            host=self.host,
            port=self.port,
            ws="websockets-sansio",
        )
        self.server = uvicorn.Server(config)

        # Start server in background with our socket
        self._server_task = asyncio.create_task(self.server.serve([self._socket]))
        logger.info("Started tool server", host=self.host, port=self.port)
        await asyncio.sleep(0.1)  # Wait for server to start properly
        # Verify server is running
        max_retries = 10
        for _ in range(max_retries):
            if self.server.started:
                break
            await asyncio.sleep(0.1)

        return ServerInfo(url=f"http://{self.host}:{self.port}", port=self.port)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Stop FastAPI server."""
        # Stop server gracefully
        if self.server:
            with contextlib.suppress(Exception):
                self.server.should_exit = True
                if hasattr(self.server, "force_exit"):
                    self.server.force_exit = True

        # Cancel server task
        if self._server_task and not self._server_task.done():
            self._server_task.cancel()
            with contextlib.suppress(asyncio.CancelledError, Exception):
                await asyncio.wait_for(self._server_task, timeout=1.0)

        # Close socket
        if self._socket:
            with contextlib.suppress(Exception):
                self._socket.close()

        # Reset state
        self.server = None
        self._server_task = None
        self._socket = None
        self.app = None


if __name__ == "__main__":
    from anyenv.code_execution.configs import LocalExecutionEnvironmentConfig

    from llmling_agent.tools.base import Tool

    def add_numbers(x: int, y: int) -> int:
        """Add two numbers."""
        return x + y

    async def main() -> None:
        tools = [Tool.from_callable(add_numbers)]
        config = LocalExecutionEnvironmentConfig()
        provider = RemoteCodeExecutor.from_tools(tools, config, server_port=9876)
        async with provider:
            result = await provider.execute_code("_result = await add_numbers(5, 3)")
            print(f"Result: {result.result}")

    asyncio.run(main())



"""Built-in event handlers for simple console output."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

from pydantic_ai import (
    FunctionToolCallEvent,
    FunctionToolResultEvent,
    PartDeltaEvent,
    PartStartEvent,
)
from pydantic_ai.messages import (
    TextPart,
    TextPartDelta,
    ThinkingPart,
    ThinkingPartDelta,
    ToolCallPart,
    ToolReturnPart,
)

from llmling_agent.agent.events import (
    FileEditProgressEvent,
    ProcessExitEvent,
    ProcessStartEvent,
    RunErrorEvent,
    StreamCompleteEvent,
    ToolCallStartEvent,
)


if TYPE_CHECKING:
    from collections.abc import Sequence

    from pydantic_ai import RunContext

    from llmling_agent.agent.events import (
        RichAgentStreamEvent,
    )
    from llmling_agent.common_types import BuiltinEventHandlerType, IndividualEventHandler


async def simple_print_handler(ctx: RunContext, event: RichAgentStreamEvent[Any]) -> None:
    """Simple event handler that prints text and basic tool information.

    Focus: Core text output and minimal tool notifications.
    Prints:
    - Text content (streaming)
    - Tool calls (name only)
    - Errors
    """
    match event:
        case (
            PartStartEvent(part=TextPart(content=delta))
            | PartDeltaEvent(delta=TextPartDelta(content_delta=delta))
        ):
            print(delta, end="", flush=True, file=sys.stderr)

        case FunctionToolCallEvent(part=ToolCallPart() as part):
            kwargs_str = ", ".join(f"{k}={v!r}" for k, v in part.args_as_dict().items())
            print(f"\n🔧 {part.tool_name}({kwargs_str})", flush=True, file=sys.stderr)

        case FunctionToolResultEvent(result=ToolReturnPart() as return_part):
            print(f"Result: {return_part.content}", file=sys.stderr)

        case RunErrorEvent(message=message):
            print(f"\n❌ Error: {message}", flush=True, file=sys.stderr)

        case StreamCompleteEvent():
            print(file=sys.stderr)  # Final newline


async def detailed_print_handler(ctx: RunContext, event: RichAgentStreamEvent[Any]) -> None:
    """Detailed event handler with rich tool execution information.

    Focus: Comprehensive execution visibility.
    Prints:
    - Text content (streaming)
    - Thinking content
    - Tool calls with inputs
    - Tool results
    - Process execution
    - File operations
    - Errors
    """
    match event:
        case (
            PartStartEvent(part=TextPart(content=delta))
            | PartDeltaEvent(delta=TextPartDelta(content_delta=delta))
        ):
            print(delta, end="", flush=True, file=sys.stderr)

        case (
            PartStartEvent(part=ThinkingPart(content=delta))
            | PartDeltaEvent(delta=ThinkingPartDelta(content_delta=delta))
        ):
            if delta:
                print(f"\n💭 {delta}", end="", flush=True, file=sys.stderr)

        case ToolCallStartEvent(tool_name=tool_name, title=title, tool_call_id=call_id):
            print(f"\n🔧 {tool_name}: {title} (#{call_id[:8]})", flush=True, file=sys.stderr)

        case FunctionToolCallEvent(part=ToolCallPart(tool_name=tool_name, args=args)):
            args_str = str(args)
            if len(args_str) > 100:  # noqa: PLR2004
                args_str = args_str[:97] + "..."
            print(f"  📝 Input: {args_str}", flush=True, file=sys.stderr)

        case FunctionToolResultEvent(result=ToolReturnPart(content=content, tool_name=tool_name)):
            result_str = str(content)
            if len(result_str) > 150:  # noqa: PLR2004
                result_str = result_str[:147] + "..."
            print(f"  ✅ {tool_name}: {result_str}", flush=True, file=sys.stderr)

        case ProcessStartEvent(process_id=pid, command=command, success=success):
            if success:
                print(f"  🖥️  Started process: {command} (PID: {pid})", flush=True, file=sys.stderr)
            else:
                print(f"  ❌ Failed to start: {command}", flush=True, file=sys.stderr)

        case ProcessExitEvent(process_id=pid, exit_code=code, success=success):
            status = "✅" if success else "❌"
            print(f"  {status} Process {pid} exited: code {code}", flush=True, file=sys.stderr)

        case FileEditProgressEvent(path=path, status=status):
            emoji = {"in_progress": "✏️", "completed": "✅", "failed": "❌"}.get(status, "📝")
            print(f"  {emoji} {status}: {path}", flush=True, file=sys.stderr)

        case RunErrorEvent(message=message, code=code):
            error_info = f" [{code}]" if code else ""
            print(f"\n❌ Error{error_info}: {message}", flush=True, file=sys.stderr)

        case StreamCompleteEvent():
            print(file=sys.stderr)  # Final newline


def resolve_event_handlers(
    event_handlers: Sequence[IndividualEventHandler | BuiltinEventHandlerType] | None,
) -> list[IndividualEventHandler]:
    """Resolve event handlers, converting builtin handler names to actual handlers."""
    if not event_handlers:
        return []
    builtin_map = {"simple": simple_print_handler, "detailed": detailed_print_handler}
    return [builtin_map[h] if isinstance(h, str) else h for h in event_handlers]



"""Models for PydanticAI builtin tools configuration."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal

from pydantic import Field
from pydantic_ai import (
    CodeExecutionTool,
    ImageGenerationTool,
    MCPServerTool,
    UrlContextTool,
    WebSearchTool,
)

from llmling_agent_config.tools import BaseToolConfig


if TYPE_CHECKING:
    from pydantic_ai.builtin_tools import AbstractBuiltinTool, MemoryTool, WebSearchUserLocation


class BaseBuiltinToolConfig(BaseToolConfig):
    """Base configuration for PydanticAI builtin tools."""

    def get_builtin_tool(self) -> AbstractBuiltinTool:
        """Convert config to PydanticAI builtin tool instance."""
        raise NotImplementedError


class WebSearchToolConfig(BaseBuiltinToolConfig):
    """Configuration for PydanticAI web search builtin tool."""

    type: Literal["web_search"] = Field("web_search", init=False)
    """Web search builtin tool."""

    search_context_size: Literal["low", "medium", "high"] = Field(
        default="medium",
        examples=["low", "medium", "high"],
        title="Search context size",
    )
    """The search context size parameter controls how much context is retrieved."""

    user_location: WebSearchUserLocation | None = Field(default=None, title="User location")
    """User location for localizing search results (city, country, region, timezone)."""

    blocked_domains: list[str] | None = Field(
        default=None,
        examples=[["spam.com", "ads.example.com"], ["blocked.site"]],
        title="Blocked domains",
    )
    """Domains that will never appear in results."""

    allowed_domains: list[str] | None = Field(
        default=None,
        examples=[["wikipedia.org", "github.com"], ["trusted.site"]],
        title="Allowed domains",
    )
    """Only these domains will be included in results."""

    max_uses: int | None = Field(default=None, examples=[5, 10, 20], title="Maximum uses")
    """Maximum number of times the tool can be used."""

    def get_builtin_tool(self) -> WebSearchTool:
        """Convert config to WebSearchTool instance."""
        return WebSearchTool(
            search_context_size=self.search_context_size,
            user_location=self.user_location,
            blocked_domains=self.blocked_domains,
            allowed_domains=self.allowed_domains,
            max_uses=self.max_uses,
        )


class CodeExecutionToolConfig(BaseBuiltinToolConfig):
    """Configuration for PydanticAI code execution builtin tool."""

    type: Literal["code_execution"] = Field("code_execution", init=False)
    """Code execution builtin tool."""

    def get_builtin_tool(self) -> CodeExecutionTool:
        """Convert config to CodeExecutionTool instance."""
        return CodeExecutionTool()


class UrlContextToolConfig(BaseBuiltinToolConfig):
    """Configuration for PydanticAI URL context builtin tool."""

    type: Literal["url_context"] = Field("url_context", init=False)
    """URL context builtin tool."""

    def get_builtin_tool(self) -> UrlContextTool:
        """Convert config to UrlContextTool instance."""
        return UrlContextTool()


class ImageGenerationToolConfig(BaseBuiltinToolConfig):
    """Configuration for PydanticAI image generation builtin tool."""

    type: Literal["image_generation"] = Field("image_generation", init=False)
    """Image generation builtin tool."""

    background: Literal["transparent", "opaque", "auto"] = Field(
        default="auto",
        examples=["transparent", "opaque", "auto"],
        title="Background type",
    )
    """Background type for the generated image."""

    input_fidelity: Literal["high", "low"] | None = Field(
        default=None,
        examples=["high", "low"],
        title="Input fidelity",
    )
    """Control how much effort the model will exert to match input image features."""

    moderation: Literal["auto", "low"] = Field(
        default="auto",
        examples=["auto", "low"],
        title="Moderation level",
    )
    """Moderation level for the generated image."""

    output_compression: int = Field(
        default=100,
        ge=0,
        le=100,
        examples=[80, 90, 100],
        title="Output compression",
    )
    """Compression level for the output image."""

    output_format: Literal["png", "webp", "jpeg"] | None = Field(
        default=None,
        examples=["png", "webp", "jpeg"],
        title="Output format",
    )
    """The output format of the generated image."""

    partial_images: int = Field(
        default=0,
        ge=0,
        examples=[0, 2, 4],
        title="Partial images count",
    )
    """Number of partial images to generate in streaming mode."""

    quality: Literal["low", "medium", "high", "auto"] = Field(
        default="auto",
        examples=["low", "medium", "high", "auto"],
        title="Image quality",
    )
    """The quality of the generated image."""

    size: Literal["1024x1024", "1024x1536", "1536x1024", "auto"] = Field(
        default="auto",
        examples=["1024x1024", "1024x1536", "1536x1024", "auto"],
        title="Image size",
    )
    """The size of the generated image."""

    def get_builtin_tool(self) -> ImageGenerationTool:
        """Convert config to ImageGenerationTool instance."""
        return ImageGenerationTool(
            background=self.background,
            input_fidelity=self.input_fidelity,
            moderation=self.moderation,
            output_compression=self.output_compression,
            output_format=self.output_format,
            partial_images=self.partial_images,
            quality=self.quality,
            size=self.size,
        )


class MemoryToolConfig(BaseBuiltinToolConfig):
    """Configuration for PydanticAI memory builtin tool."""

    type: Literal["memory"] = Field("memory", init=False)
    """Memory builtin tool."""

    def get_builtin_tool(self) -> MemoryTool:
        """Convert config to MemoryTool instance."""
        from pydantic_ai import MemoryTool

        return MemoryTool()


class MCPServerToolConfig(BaseBuiltinToolConfig):
    """Configuration for PydanticAI MCP server builtin tool."""

    type: Literal["mcp_server"] = Field("mcp_server", init=False)
    """MCP server builtin tool."""

    server_id: str = Field(
        alias="id",
        examples=["my_mcp_server", "code_tools", "web_api"],
        title="Server ID",
    )
    """A unique identifier for the MCP server."""

    url: str = Field(
        examples=["https://api.example.com/mcp", "http://localhost:8080"],
        title="Server URL",
    )
    """The URL of the MCP server to use."""

    authorization_token: str | None = Field(
        default=None,
        examples=["Bearer token123", "api_key_abc"],
        title="Authorization token",
    )
    """Authorization header to use when making requests to the MCP server."""

    description: str | None = Field(
        default=None,
        examples=["External API tools", "Code execution server"],
        title="Server description",
    )
    """A description of the MCP server."""

    allowed_tools: list[str] | None = Field(
        default=None,
        examples=[["search", "fetch"], ["execute", "compile"]],
        title="Allowed tools",
    )
    """A list of tools that the MCP server can use."""

    headers: dict[str, str] | None = Field(default=None, title="HTTP headers")
    """Optional HTTP headers to send to the MCP server."""

    def get_builtin_tool(self) -> MCPServerTool:
        """Convert config to MCPServerTool instance."""
        return MCPServerTool(
            id=self.server_id,
            url=self.url,
            authorization_token=self.authorization_token,
            description=self.description,
            allowed_tools=self.allowed_tools,
            headers=self.headers,
        )


# Union type for builtin tool configs
BuiltinToolConfig = Annotated[
    WebSearchToolConfig
    | CodeExecutionToolConfig
    | UrlContextToolConfig
    | ImageGenerationToolConfig
    | MemoryToolConfig
    | MCPServerToolConfig,
    Field(discriminator="type"),
]



"""Helper functions for running examples in different environments."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Self


if TYPE_CHECKING:
    from collections.abc import Awaitable, Iterator


EXAMPLES_DIR = Path("src/llmling_agent_docs/examples")


def is_pyodide() -> bool:
    """Check if code is running in a Pyodide environment."""
    try:
        from js import Object  # type: ignore[import-not-found] # noqa: F401

        return True  # noqa: TRY300
    except ImportError:
        return False


def get_config_path(module_path: str | None = None, filename: str = "config.yml") -> Path:
    """Get the configuration file path based on environment.

    Args:
        module_path: Optional __file__ from the calling module (ignored in Pyodide)
        filename: Name of the config file (default: config.yml)

    Returns:
        Path to the configuration file
    """
    if is_pyodide():
        return Path(filename)
    if module_path is None:
        raise ValueError("module_path is required in non-Pyodide environment")
    return Path(module_path).parent / filename


def run[T](coro: Awaitable[T]) -> T:
    """Run a coroutine in both normal Python and Pyodide environments."""
    try:
        # Check if we're in an event loop
        asyncio.get_running_loop()
        # If we are, run until complete
        return asyncio.get_event_loop().run_until_complete(coro)
    except RuntimeError:
        # No running event loop, create one
        return asyncio.run(coro)  # type: ignore


@dataclass
class Example:
    """Represents a LLMling-agent example with its metadata."""

    name: str
    path: Path
    title: str
    description: str
    icon: str = "octicon:code-16"

    @property
    def files(self) -> list[Path]:
        """Get all Python and YAML files (excluding __init__.py)."""
        return [
            f
            for f in self.path.glob("**/*.*")
            if f.suffix in {".py", ".yml"} and not f.name.startswith("__")
        ]

    @property
    def docs(self) -> Path | None:
        """Get docs.md file if it exists."""
        docs = self.path / "docs.md"
        return docs if docs.exists() else None

    @classmethod
    def from_directory(cls, path: Path) -> Self | None:
        """Create Example from directory if it's a valid example."""
        if not path.is_dir() or path.name.startswith("__"):
            return None

        init_file = path / "__init__.py"
        if not init_file.exists():
            return None

        # Load the module to get variables
        namespace: dict[str, str] = {}
        with init_file.open() as f:
            exec(f.read(), namespace)

        # Get metadata with defaults
        title = namespace.get("TITLE", path.name.replace("_", " ").title())
        icon = namespace.get("ICON", "octicon:code-16")
        description = namespace.get("__doc__", "")

        return cls(
            name=path.name,
            path=path,
            title=title,
            description=description,
            icon=icon,
        )


def iter_examples(root: Path | str | None = None) -> Iterator[Example]:
    """Iterate over all available examples.

    Args:
        root: Optional root directory (defaults to llmling_agent_docs/examples)
    """
    root = Path(root) if root else EXAMPLES_DIR

    for path in sorted(root.iterdir()):
        if example := Example.from_directory(path):
            yield example


def get_example(name: str, root: Path | str | None = None) -> Example:
    """Get a specific example by name."""
    for example in iter_examples(root):
        if example.name == name:
            return example
    raise KeyError(f"Example {name!r} not found")


if __name__ == "__main__":
    # Example usage:
    for ex in iter_examples():
        print(f"\n{ex.title} ({ex.name})")
        print(f"Icon: {ex.icon}")
        print(f"Files: {len(ex.files)}")
        if ex.docs:
            print("Has docs.md")
        print(f"Description: {ex.description.splitlines()[0]}")



"""Content types for messages."""

from __future__ import annotations

import base64
from typing import TYPE_CHECKING, Annotated, Literal, Self

from pydantic import ConfigDict, Field
from pydantic_ai import AudioUrl, BinaryContent, BinaryImage, DocumentUrl, ImageUrl
from schemez import Schema
from upathtools import read_path, to_upath


if TYPE_CHECKING:
    from pydantic_ai import UserContent
    from upath.types import JoinablePathLike


DetailLevel = Literal["high", "low", "auto"]


class BaseContent(Schema):
    """Base class for special content types (non-text)."""

    type: str = Field(init=False)
    """Discriminator field for content types."""

    description: str | None = None
    """Human-readable description of the content."""

    model_config = ConfigDict(frozen=True)

    def to_pydantic_ai(self) -> UserContent:
        msg = "Subclasses must implement this method"
        raise NotImplementedError(msg)


class BaseImageContent(BaseContent):
    """Base for image content."""

    detail: DetailLevel | None = None
    """Detail level for image processing by vision models.
    - high: Maximum resolution (up to 2048x2048)
    - low: Lower resolution (512x512)
    - auto: Let model decide based on content
    """

    @classmethod
    async def from_path(
        cls,
        path: JoinablePathLike,
        *,
        detail: DetailLevel | None = None,
        description: str | None = None,
    ) -> ImageURLContent | ImageBase64Content:
        """Create image content from any path.

        Automatically chooses between URL and base64 based on path type.
        Downloads and converts remote content if needed.

        Args:
            path: Local path or URL to image
            detail: Optional detail level for processing
            description: Optional description of the image
        """
        path_obj = to_upath(path)
        # For http(s) URLs, pass through as URL content
        if path_obj.protocol in {"http", "https"}:
            return ImageURLContent(url=str(path_obj), detail=detail, description=description)

        # For all other paths, read and convert to base64
        data = await read_path(path_obj, mode="rb")
        content = base64.b64encode(data).decode()
        return ImageBase64Content(data=content, detail=detail, description=description)


class ImageURLContent(BaseImageContent):
    """Image from URL."""

    type: Literal["image_url"] = Field("image_url", init=False)
    """URL-based image."""

    url: str
    """URL to the image."""

    def to_pydantic_ai(self) -> ImageUrl:
        return ImageUrl(url=self.url)


class ImageBase64Content(BaseImageContent):
    """Image from base64 data."""

    type: Literal["image_base64"] = Field("image_base64", init=False)
    """Base64-encoded image."""

    data: str
    """Base64-encoded image data."""

    mime_type: str = "image/jpeg"
    """MIME type of the image."""

    def to_pydantic_ai(self) -> BinaryImage:
        binary_data = base64.b64decode(self.data)
        return BinaryImage(data=binary_data, media_type=self.mime_type or "image/jpeg")

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        *,
        detail: DetailLevel | None = None,
        description: str | None = None,
    ) -> ImageBase64Content:
        """Create image content from raw bytes.

        Args:
            data: Raw image bytes
            detail: Optional detail level for processing
            description: Optional description of the image
        """
        content = base64.b64encode(data).decode()
        return cls(data=content, detail=detail, description=description)


class BasePDFContent(BaseContent):
    """Base for PDF document content."""

    detail: DetailLevel | None = None
    """Detail level for document processing by models."""

    @classmethod
    async def from_path(
        cls,
        path: JoinablePathLike,
        *,
        detail: DetailLevel | None = None,
        description: str | None = None,
    ) -> PDFURLContent | PDFBase64Content:
        """Create PDF content from any path.

        Args:
            path: Local path or URL to PDF
            detail: Optional detail level for processing
            description: Optional description of the document
        """
        from upathtools import to_upath

        path_obj = to_upath(path)

        # For http(s) URLs, pass through as URL content
        if path_obj.protocol in {"http", "https"}:
            return PDFURLContent(url=str(path_obj), detail=detail, description=description)

        # For all other paths, read and convert to base64
        data = await read_path(path_obj, mode="rb")
        content = base64.b64encode(data).decode()
        return PDFBase64Content(data=content, detail=detail, description=description)


class PDFURLContent(BasePDFContent):
    """PDF from URL."""

    type: Literal["pdf_url"] = Field("pdf_url", init=False)
    """URL-based PDF."""

    url: str
    """URL to the PDF document."""

    def to_pydantic_ai(self) -> DocumentUrl:
        return DocumentUrl(url=self.url)


class PDFBase64Content(BasePDFContent):
    """PDF from base64 data."""

    type: Literal["pdf_base64"] = Field("pdf_base64", init=False)
    """Base64-data based PDF."""

    data: str
    """Base64-encoded PDF data."""

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        *,
        detail: DetailLevel | None = None,
        description: str | None = None,
    ) -> Self:
        """Create PDF content from raw bytes."""
        content = base64.b64encode(data).decode()
        return cls(data=content, detail=detail, description=description)

    def to_pydantic_ai(self) -> BinaryContent:
        binary_data = base64.b64decode(self.data)
        return BinaryContent(binary_data, media_type="application/pdf")


class AudioContent(BaseContent):
    """Base for audio content."""

    format: str | None = None  # mp3, wav, etc
    """Audio format."""

    description: str | None = None
    """Human-readable description of the content."""


class AudioURLContent(AudioContent):
    """Audio from URL."""

    type: Literal["audio_url"] = Field("audio_url", init=False)
    """URL-based audio."""

    url: str
    """URL to the audio."""

    def to_pydantic_ai(self) -> AudioUrl:
        return AudioUrl(url=self.url)


class AudioBase64Content(AudioContent):
    """Audio from base64 data."""

    type: Literal["audio_base64"] = Field("audio_base64", init=False)
    """Base64-encoded audio."""

    data: str
    """Audio data in base64 format."""

    format: str | None = None  # mp3, wav, etc
    """Audio format."""

    def to_pydantic_ai(self) -> BinaryContent:
        binary_data = base64.b64decode(self.data)
        return BinaryContent(data=binary_data, media_type=f"audio/{self.format or 'mp3'}")

    @classmethod
    def from_bytes(cls, data: bytes, audio_format: str = "mp3") -> Self:
        """Create from raw bytes."""
        return cls(data=base64.b64encode(data).decode(), format=audio_format)

    @classmethod
    def from_path(cls, path: JoinablePathLike) -> Self:
        """Create from file path with auto format detection."""
        import mimetypes

        path_obj = to_upath(path)
        mime_type, _ = mimetypes.guess_type(str(path_obj))
        fmt = (
            mime_type.removeprefix("audio/")
            if mime_type and mime_type.startswith("audio/")
            else "mp3"
        )

        return cls(data=base64.b64encode(path_obj.read_bytes()).decode(), format=fmt)

    @property
    def mime_type(self) -> str:
        """Return the MIME type of the audio."""
        return f"audio/{self.format or 'mp3'}"


class VideoContent(BaseContent):
    """Base for video content."""

    format: str | None = None
    """Video format."""

    description: str | None = None
    """Human-readable description of the content."""


class VideoURLContent(VideoContent):
    """Video from URL."""

    type: Literal["video_url"] = Field("video_url", init=False)
    """URL-based video."""

    url: str
    """URL to the video."""


Content = Annotated[
    ImageURLContent
    | ImageBase64Content
    | PDFURLContent
    | PDFBase64Content
    | AudioURLContent
    | AudioBase64Content
    | VideoURLContent,
    Field(discriminator="type"),
]



"""Provider for skills tools."""

from __future__ import annotations

from llmling_agent.agent.context import AgentContext  # noqa: TC001
from llmling_agent.resource_providers import StaticResourceProvider


BASE_DESC = """Load a Claude Code Skill and return its instructions.

This tool provides access to Claude Code Skills - specialized workflows and techniques
for handling specific types of tasks. When you need to use a skill, call this tool
with the skill name.

Available skills:"""


async def load_skill(ctx: AgentContext, skill_name: str) -> str:
    """Load a Claude Code Skill and return its instructions.

    Args:
        ctx: Agent context providing access to pool and skills
        skill_name: Name of the skill to load

    Returns:
        The full skill instructions for execution
    """
    if not ctx.pool:
        return "No agent pool available - skills require pool context"

    skills_manager = ctx.pool.skills
    skills = skills_manager.list_skills()

    if not skills:
        return "No skills available in the pool"

    skill = next((s for s in skills if s.name == skill_name), None)
    if not skill:
        available = ", ".join(s.name for s in skills)
        return f"Skill {skill_name!r} not found. Available skills: {available}"

    try:
        instructions = skills_manager.get_skill_instructions(skill_name)
    except Exception as e:  # noqa: BLE001
        return f"Failed to load skill {skill_name!r}: {e}"
    return f"""# {skill.name}

{instructions}

---
Skill directory: {skill.skill_path}
"""


async def list_skills(ctx: AgentContext) -> str:
    """List all available skills.

    Returns:
        Formatted list of available skills with descriptions
    """
    if not ctx.pool:
        return "No agent pool available - skills require pool context"

    skills = ctx.pool.skills.list_skills()

    if not skills:
        return "No skills available"

    lines = ["Available skills:", ""]
    lines.extend(f"- **{skill.name}**: {skill.description}" for skill in skills)
    return "\n".join(lines)


class SkillsTools(StaticResourceProvider):
    """Provider for Claude Code Skills tools.

    Provides tools to discover and load skills from the pool's skills registry.
    Skills are discovered from configured directories (e.g., ~/.claude/skills/,
    .claude/skills/).

    The pool manages skill discovery; this toolset just provides access to them.
    """

    def __init__(self, name: str = "skills") -> None:
        super().__init__(name=name)
        self._tools = [
            self.create_tool(
                load_skill,
                category="read",
                read_only=True,
                idempotent=True,
            ),
            self.create_tool(
                list_skills,
                category="read",
                read_only=True,
                idempotent=True,
            ),
        ]



"""Storage package."""

from llmling_agent.storage.manager import StorageManager
from llmling_agent.storage.serialization import (
    deserialize_messages,
    deserialize_parts,
    serialize_messages,
    serialize_parts,
)

__all__ = [
    "StorageManager",
    "deserialize_messages",
    "deserialize_parts",
    "serialize_messages",
    "serialize_parts",
]



"""Slash command wrapper for Agent that injects command events into streams."""

from __future__ import annotations

import asyncio
import re
from typing import TYPE_CHECKING, Any, cast

from slashed.events import (
    CommandExecutedEvent,
    CommandOutputEvent as SlashedCommandOutputEvent,
)

from llmling_agent.agent.events import CommandCompleteEvent, CommandOutputEvent
from llmling_agent.log import get_logger


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable

    from slashed import CommandContext, CommandStore
    from slashed.events import CommandStoreEvent

    from llmling_agent.agent import Agent, AGUIAgent
    from llmling_agent.agent.acp_agent import ACPAgent
    from llmling_agent.agent.events import SlashedAgentStreamEvent
    from llmling_agent.common_types import PromptCompatible


logger = get_logger(__name__)
SLASH_PATTERN = re.compile(r"^/([\w-]+)(?:\s+(.*))?$")


def _parse_slash_command(command_text: str) -> tuple[str, str] | None:
    """Parse slash command into name and args.

    Args:
        command_text: Full command text

    Returns:
        Tuple of (cmd_name, args) or None if invalid
    """
    if match := SLASH_PATTERN.match(command_text.strip()):
        cmd_name = match.group(1)
        args = match.group(2) or ""
        return cmd_name, args.strip()
    return None


class SlashedAgent[TDeps, OutputDataT]:
    """Wrapper around Agent that handles slash commands in streams.

    Uses the "commands first" strategy from the ACP adapter:
    1. Execute all slash commands first
    2. Then process remaining content through wrapped agent
    3. If only commands, end without LLM processing
    """

    def __init__(
        self,
        agent: Agent[TDeps, OutputDataT] | ACPAgent | AGUIAgent,
        command_store: CommandStore | None = None,
        *,
        context_data_factory: Callable[[], Any] | None = None,
    ) -> None:
        """Initialize with wrapped agent and command store.

        Args:
            agent: The agent to wrap
            command_store: Command store for slash commands (creates default if None)
            context_data_factory: Optional factory for creating command context data
        """
        self.agent = agent
        self._context_data_factory = context_data_factory
        self._event_queue: asyncio.Queue[CommandStoreEvent] | None = None

        # Create store with our streaming event handler
        if command_store is None:
            from slashed import CommandStore

            from llmling_agent_commands import get_commands

            cmds = get_commands()
            self.command_store = CommandStore(event_handler=self._emit_event, commands=cmds)
        else:
            self.command_store = command_store

    async def _emit_event(self, event: CommandStoreEvent) -> None:
        """Bridge store events to async queue during command execution."""
        if self._event_queue:
            await self._event_queue.put(event)

    def _is_slash_command(self, text: str) -> bool:
        """Check if text starts with a slash command.

        Args:
            text: Text to check

        Returns:
            True if text is a slash command
        """
        return bool(SLASH_PATTERN.match(text.strip()))

    async def _execute_slash_command_streaming(
        self, command_text: str
    ) -> AsyncGenerator[CommandOutputEvent | CommandCompleteEvent]:
        """Execute a single slash command and yield events as they happen.

        Args:
            command_text: Full command text including slash

        Yields:
            Command output and completion events
        """
        parsed = _parse_slash_command(command_text)
        if not parsed:
            logger.warning("Invalid slash command", command=command_text)
            yield CommandCompleteEvent(command="unknown", success=False)
            return

        cmd_name, args = parsed

        # Set up event queue for this command execution
        self._event_queue = asyncio.Queue()
        context_data = (  # Create command context
            self._context_data_factory() if self._context_data_factory else self.agent.context
        )

        cmd_ctx = self.command_store.create_context(data=context_data)
        command_str = f"{cmd_name} {args}".strip()
        execute_task = asyncio.create_task(self.command_store.execute_command(command_str, cmd_ctx))

        success = True
        try:
            # Yield events from queue as command runs
            while not execute_task.done():
                try:
                    # Wait for events with short timeout to check task completion
                    event = await asyncio.wait_for(self._event_queue.get(), timeout=0.1)
                    # Convert store events to our stream events
                    match event:
                        case SlashedCommandOutputEvent(output=output):
                            yield CommandOutputEvent(command=cmd_name, output=output)
                        case CommandExecutedEvent(success=False, error=error) if error:
                            output = f"Command error: {error}"
                            yield CommandOutputEvent(command=cmd_name, output=output)
                            success = False
                except TimeoutError:
                    continue

            # Ensure command task completes and handle any remaining events
            try:
                await execute_task
            except Exception as e:
                logger.exception("Command execution failed", command=cmd_name)
                success = False
                yield CommandOutputEvent(command=cmd_name, output=f"Command error: {e}")

            # Drain any remaining events from queue
            while not self._event_queue.empty():
                try:
                    match self._event_queue.get_nowait():
                        case SlashedCommandOutputEvent(output=output):
                            yield CommandOutputEvent(command=cmd_name, output=output)
                except asyncio.QueueEmpty:
                    break

            # Always yield completion event
            yield CommandCompleteEvent(command=cmd_name, success=success)

        finally:
            # Clean up event queue
            self._event_queue = None

    async def run_stream(
        self,
        *prompts: PromptCompatible,
        **kwargs: Any,
    ) -> AsyncGenerator[SlashedAgentStreamEvent[OutputDataT]]:
        """Run agent with slash command support.

        Separates slash commands from regular prompts, executes commands first,
        then processes remaining content through the wrapped agent.

        Args:
            *prompts: Input prompts (may include slash commands)
            **kwargs: Additional arguments passed to agent.run_stream

        Yields:
            Stream events from command execution and agent processing
        """
        # Separate slash commands from regular content
        commands: list[str] = []
        regular_prompts: list[Any] = []

        for prompt in prompts:
            if isinstance(prompt, str) and self._is_slash_command(prompt):
                logger.debug("Found slash command", command=prompt)
                commands.append(prompt.strip())
            else:
                regular_prompts.append(prompt)

        # Execute all commands first with streaming
        if commands:
            for command in commands:
                logger.info("Processing slash command", command=command)
                async for cmd_event in self._execute_slash_command_streaming(command):
                    yield cmd_event

        # If we have regular content, process it through the agent
        if regular_prompts:
            logger.debug("Processing prompts through agent", num_prompts=len(regular_prompts))
            async for event in self.agent.run_stream(*regular_prompts, **kwargs):
                # ACPAgent always returns str, cast to match OutputDataT
                yield cast("SlashedAgentStreamEvent[OutputDataT]", event)

        # If we only had commands and no regular content, we're done
        # (no additional events needed)

    def __getattr__(self, name: str) -> Any:
        """Delegate attribute access to wrapped agent.

        Args:
            name: Attribute name

        Returns:
            Attribute value from wrapped agent
        """
        return getattr(self.agent, name)


if __name__ == "__main__":
    import asyncio

    from llmling_agent import Agent

    async def main() -> None:
        agent = Agent("test-agent", model="test", session=False)
        slashed = SlashedAgent(agent)  # Uses built-in commands by default

        # Add a simple test command that outputs multiple lines
        @slashed.command_store.command(name="test-streaming", category="test")
        async def test_streaming(ctx: CommandContext[Any], *args: Any, **kwargs: Any) -> None:
            """Test command that outputs multiple lines."""
            await ctx.print("Starting streaming test...")
            for i in range(3):
                await ctx.print(f"Output line {i + 1}")
                await asyncio.sleep(0.1)  # Small delay to simulate work
            await ctx.print("Streaming test complete!")

        print("Testing SlashedAgent streaming:")
        async for event in slashed.run_stream("/test-streaming"):
            print(f"Event: {event}")

    asyncio.run(main())



"""MCP protocol server implementation."""

from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any

from fastmcp import FastMCP
from fastmcp.server.middleware.caching import ResponseCachingMiddleware
from key_value.aio.stores.disk import DiskStore
from mcp import CreateMessageResult
from mcp.types import TextContent
import platformdirs

import llmling_agent
from llmling_agent.log import get_logger
from llmling_agent_server import BaseServer
from llmling_agent_server.mcp_server.handlers import register_handlers


if TYPE_CHECKING:
    from collections.abc import Callable
    from contextlib import AbstractAsyncContextManager
    from typing import Any

    from fastmcp import SamplingMessage, ServerSession
    import mcp
    from mcp.server.lowlevel.server import LifespanResultT
    from mcp.shared.context import LifespanContextT, RequestContext
    from mcp.types import CreateMessageRequestParams as SamplingParams

    from llmling_agent import AgentPool
    from llmling_agent_config.pool_server import MCPPoolServerConfig

    LifespanHandler = Callable[
        [FastMCP[LifespanResultT]],
        AbstractAsyncContextManager[LifespanResultT],
    ]

logger = get_logger(__name__)


llmling_dir = platformdirs.user_config_dir("llmling-agent")

store = DiskStore(directory=llmling_dir)
middleware = ResponseCachingMiddleware(cache_storage=store)


class MCPServer(BaseServer):
    """MCP protocol server implementation."""

    def __init__(
        self,
        pool: AgentPool[Any],
        config: MCPPoolServerConfig,
        lifespan: (LifespanHandler | None) = None,
        instructions: str | None = None,
        name: str = "llmling-server",
        raise_exceptions: bool = False,
    ) -> None:
        """Initialize server with agent pool.

        Args:
            pool: AgentPool to expose through MCP
            config: Server configuration
            name: Server name for MCP protocol
            lifespan: Lifespan context manager
            instructions: Instructions for Server usage
            raise_exceptions: Whether to raise exceptions during server start
        """
        from llmling_agent.resource_providers.pool import PoolResourceProvider

        super().__init__(pool, name=name, raise_exceptions=raise_exceptions)
        self.provider = PoolResourceProvider(pool, zed_mode=config.zed_mode)
        self.config = config

        # Handle Zed mode if enabled
        if config.zed_mode:
            pass
            # TODO: adapt zed wrapper to work with ResourceProvider
            # prepare_runtime_for_zed(runtime)

        self._subscriptions: defaultdict[str, set[mcp.ServerSession]] = defaultdict(set)

        self.fastmcp = FastMCP(
            instructions=instructions,
            lifespan=lifespan,
            version=llmling_agent.__version__,
            middleware=[middleware],
            # sampling_handler=self._sampling_handler,
        )
        self.server = self.fastmcp._mcp_server
        register_handlers(self)

    async def _sampling_handler(
        self,
        messages: list[SamplingMessage],
        params: SamplingParams,
        context: RequestContext[ServerSession, LifespanContextT],
    ) -> CreateMessageResult:
        # This is a fallback handler in case the client does not support sampling.
        return CreateMessageResult(
            role="user",
            content=TextContent(type="text", text="test"),
            model="test",
        )

    async def _start_async(self) -> None:
        """Start the server (blocking async - runs until stopped)."""
        if self.config.transport == "stdio":
            await self.fastmcp.run_async(transport=self.config.transport)
        else:
            await self.fastmcp.run_async(
                transport=self.config.transport,
                host=self.config.host,
                port=self.config.port,
            )

    @property
    def current_session(self) -> mcp.ServerSession:
        """Get current session from request context."""
        try:
            return self.server.request_context.session
        except LookupError as exc:
            msg = "No active request context"
            raise RuntimeError(msg) from exc

    async def report_progress(
        self,
        progress: float,
        total: float | None = None,
        message: str | None = None,
        related_request_id: str | None = None,
    ) -> None:
        """Report progress for the current operation."""
        progress_token = (
            self.server.request_context.meta.progressToken
            if self.server.request_context.meta
            else None
        )

        if progress_token is None:
            return

        await self.server.request_context.session.send_progress_notification(
            progress_token=progress_token,
            progress=progress,
            total=total,
            message=message,
            related_request_id=related_request_id,
        )

    @property
    def client_info(self) -> mcp.Implementation | None:
        """Get client info from current session."""
        session = self.current_session
        if not session.client_params:
            return None
        return session.client_params.clientInfo

    async def notify_tool_list_changed(self) -> None:
        """Notify clients about tool list changes."""
        try:
            self.task_manager.create_task(self.current_session.send_tool_list_changed())
        except RuntimeError:
            self.log.debug("No active session for notification")
        except Exception:
            self.log.exception("Failed to send tool list change notification")

    async def notify_prompt_list_changed(self) -> None:
        """Notify clients about prompt list changes."""
        try:
            self.task_manager.create_task(self.current_session.send_prompt_list_changed())
        except RuntimeError:
            self.log.debug("No active session for notification")
        except Exception:
            self.log.exception("Failed to send prompt list change notification")



"""ACP-based input provider for agent interactions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any
import webbrowser

from mcp import types

from acp import PermissionOption
from acp.utils import DEFAULT_PERMISSION_OPTIONS
from llmling_agent.log import get_logger
from llmling_agent.ui.base import InputProvider


if TYPE_CHECKING:
    from acp import RequestPermissionResponse
    from llmling_agent.agent.context import ConfirmationResult
    from llmling_agent.messaging import ChatMessage
    from llmling_agent.messaging.context import NodeContext
    from llmling_agent.tools.base import Tool
    from llmling_agent_server.acp_server.session import ACPSession

logger = get_logger(__name__)


def _create_enum_elicitation_options(schema: dict[str, Any]) -> list[PermissionOption] | None:
    """Create permission options for enum elicitation.

    max 4 options (not more ids available)
    """
    enum_values = schema.get("enum", [])
    enum_names = schema.get("enumNames", enum_values)  # Use enumNames if available
    # Limit to 3 choices + cancel to keep ACP UI reasonable
    if len(enum_values) > 3:  # noqa: PLR2004
        logger.warning("Enum truncated for UI", enum_count=len(enum_values), showing_count=3)
        enum_values = enum_values[:3]
        enum_names = enum_names[:3]

    if not enum_values:
        return None

    options = []
    for i, (value, name) in enumerate(zip(enum_values, enum_names, strict=False)):
        opt_id = f"enum_{i}_{value}"
        option = PermissionOption(option_id=opt_id, name=str(name), kind="allow_once")
        options.append(option)
    # Add cancel option
    options.append(PermissionOption(option_id="cancel", name="Cancel", kind="reject_always"))

    return options


def _is_boolean_schema(schema: dict[str, Any]) -> bool:
    """Check if the elicitation schema is requesting a boolean value."""
    # Direct boolean type
    if schema.get("type") == "boolean":
        return True

    # Check if it's an object with a single boolean property (common pattern)
    if schema.get("type") == "object":
        properties = schema.get("properties", {})
        if len(properties) == 1:
            prop_schema = next(iter(properties.values()))
            return bool(prop_schema.get("type") == "boolean")

    return False


def _is_enum_schema(schema: dict[str, Any]) -> bool:
    """Check if the elicitation schema is requesting an enum value."""
    return schema.get("type") == "string" and "enum" in schema


def _create_boolean_elicitation_options() -> list[PermissionOption]:
    """Create permission options for boolean elicitation (Yes/No)."""
    return [
        PermissionOption(option_id="true", name="Yes", kind="allow_once"),
        PermissionOption(option_id="false", name="No", kind="reject_once"),
        PermissionOption(option_id="cancel", name="Cancel", kind="reject_always"),
    ]


class ACPInputProvider(InputProvider):
    """Input provider that uses ACP session for user interactions.

    This provider enables tool confirmation and elicitation requests
    through the ACP protocol, allowing clients to interact with agents
    for permission requests and additional input.
    """

    def __init__(self, session: ACPSession) -> None:
        """Initialize ACP input provider.

        Args:
            session: Active ACP session for handling requests
        """
        self.session = session
        self._tool_approvals: dict[str, str] = {}  #  tool_name -> "allow_always" | "reject_always"

    async def get_tool_confirmation(
        self,
        context: NodeContext[Any],
        tool: Tool,
        args: dict[str, Any],
        message_history: list[ChatMessage[Any]] | None = None,
    ) -> ConfirmationResult:
        """Get tool execution confirmation via ACP request permission.

        Uses the ACP session's request_permission mechanism to ask
        the client for confirmation before executing the tool.

        Args:
            context: Current node context
            tool: Information about the tool to be executed
            args: Tool arguments that will be passed to the tool
            message_history: Optional conversation history

        Returns:
            Confirmation result indicating whether to allow, skip, or abort
        """
        try:
            # Check if we have a standing approval/rejection for this tool
            if tool.name in self._tool_approvals:
                standing_decision = self._tool_approvals[tool.name]
                if standing_decision == "allow_always":
                    logger.debug("Auto-allowing tool", tool_name=tool.name, reason="allow_always")
                    return "allow"
                if standing_decision == "reject_always":
                    logger.debug("Auto-rejecting tool", tool_name=tool.name, reason="reject_always")
                    return "skip"

            # Create a descriptive title for the permission request
            args_str = ", ".join(f"{k}={v}" for k, v in args.items())
            response = await self.session.requests.request_permission(
                tool_call_id=f"{tool.name}_{hash(frozenset(args.items()))}",
                title=f"Execute tool {tool.name!r} with args: {args_str}",
                options=DEFAULT_PERMISSION_OPTIONS,
            )
            # Map ACP permission response to our confirmation result
            if response.outcome.outcome == "selected":
                return self._handle_permission_response(response.outcome.option_id, tool.name)
            if response.outcome.outcome == "cancelled":
                return "skip"
            # Handle other outcomes
            logger.warning("Unexpected permission outcome", outcome=response.outcome.outcome)

        except Exception:
            logger.exception("Failed to get tool confirmation")
            # Default to abort on error to be safe
            return "abort_run"
        else:
            return "abort_run"

    def _handle_permission_response(self, option_id: str, tool_name: str) -> ConfirmationResult:
        """Handle permission response and update tool approval state."""
        match option_id:
            case "allow-once":
                return "allow"
            case "allow-always":
                self._tool_approvals[tool_name] = "allow_always"
                logger.info("Tool approval set", tool_name=tool_name, approval="allow_always")
                return "allow"
            case "reject-once":
                return "skip"
            case "reject-always":
                self._tool_approvals[tool_name] = "reject_always"
                logger.info("Tool approval set", tool_name=tool_name, approval="reject_always")
                return "skip"
            case _:
                logger.warning("Unknown permission option", option_id=option_id)
                return "abort_run"

    async def get_elicitation(  # noqa: PLR0911
        self,
        params: types.ElicitRequestParams,
    ) -> types.ElicitResult | types.ErrorData:
        """Get user response to elicitation request with basic schema support.

        Currently supports boolean schemas via ACP permission options.
        Other schemas fall back to accept/decline options.

        Args:
            params: MCP elicit request parameters

        Returns:
            Elicit result with user's response or error data
        """
        try:
            # Handle URL mode elicitation (OAuth, credentials, payments)
            if isinstance(params, types.ElicitRequestURLParams):
                msg = "URL elicitation request"
                elicit_id = params.elicitationId
                logger.info(msg, message=params.message, url=params.url, elicitation_id=elicit_id)
                tool_call_id = f"elicit_url_{elicit_id}"
                title = f"URL Authorization: {params.message}"
                url_options = [
                    PermissionOption(option_id="accept", name="Open URL", kind="allow_once"),
                    PermissionOption(option_id="decline", name="Decline", kind="reject_once"),
                ]
                response = await self.session.requests.request_permission(
                    tool_call_id=tool_call_id,
                    title=title,
                    options=url_options,
                )
                if response.outcome.outcome == "selected":
                    if response.outcome.option_id == "accept":
                        webbrowser.open(params.url)
                        return types.ElicitResult(action="accept")
                    return types.ElicitResult(action="decline")
                return types.ElicitResult(action="cancel")

            # Form mode elicitation
            schema = params.requestedSchema
            logger.info("Elicitation request", message=params.message, schema=schema)
            tool_call_id = f"elicit_{hash(params.message)}"
            title = f"Elicitation: {params.message}"

            if _is_boolean_schema(schema):
                options: list[PermissionOption] | None = _create_boolean_elicitation_options()
                response = await self.session.requests.request_permission(
                    tool_call_id=tool_call_id,
                    title=title,
                    options=options,
                )
                return self._handle_boolean_elicitation_response(response, schema)
            if _is_enum_schema(schema) and (options := _create_enum_elicitation_options(schema)):
                response = await self.session.requests.request_permission(
                    tool_call_id=tool_call_id,
                    title=title,
                    options=options,
                )
                return _handle_enum_elicitation_response(response, schema)

            options = [
                PermissionOption(option_id="accept", name="Accept", kind="allow_once"),
                PermissionOption(option_id="decline", name="Decline", kind="reject_once"),
            ]
            response = await self.session.requests.request_permission(
                tool_call_id=tool_call_id,
                title=title,
                options=options,
            )

            # Convert permission response to elicitation result
            if response.outcome.outcome == "selected":
                if response.outcome.option_id == "accept":
                    # For non-boolean schemas, return empty content
                    return types.ElicitResult(action="accept", content={})
                return types.ElicitResult(action="decline")
            if response.outcome.outcome == "cancelled":
                return types.ElicitResult(action="cancel")
            return types.ElicitResult(action="cancel")

        except Exception as e:
            logger.exception("Failed to handle elicitation")
            return types.ErrorData(code=types.INTERNAL_ERROR, message=f"Elicitation failed: {e}")

    def _handle_boolean_elicitation_response(  # noqa: PLR0911
        self, response: RequestPermissionResponse, schema: dict[str, Any]
    ) -> types.ElicitResult | types.ErrorData:
        """Handle ACP response for boolean elicitation."""
        if response.outcome.outcome == "selected":
            if response.outcome.option_id == "true":
                # Check if we need to wrap in object structure
                if schema.get("type") == "object":
                    properties = schema.get("properties", {})
                    if len(properties) == 1:
                        prop_name = next(iter(properties.keys()))
                        return types.ElicitResult(action="accept", content={prop_name: True})
                return types.ElicitResult(action="accept", content={"value": True})
            if response.outcome.option_id == "false":
                # Check if we need to wrap in object structure
                if schema.get("type") == "object":
                    properties = schema.get("properties", {})
                    if len(properties) == 1:
                        prop_name = next(iter(properties.keys()))
                        return types.ElicitResult(action="accept", content={prop_name: False})
                return types.ElicitResult(action="accept", content={"value": False})
            if response.outcome.option_id == "cancel":
                return types.ElicitResult(action="cancel")

        # Handle cancelled outcome
        if response.outcome.outcome == "cancelled":
            return types.ElicitResult(action="cancel")

        return types.ElicitResult(action="cancel")

    def clear_tool_approvals(self) -> None:
        """Clear all stored tool approval decisions.

        This resets the "allow_always" and "reject_always" states,
        so tools will ask for permission again.
        """
        approval_count = len(self._tool_approvals)
        self._tool_approvals.clear()
        logger.info("Cleared tool approval decisions", count=approval_count)

    def get_tool_approval_state(self) -> dict[str, str]:
        """Get current tool approval state for debugging/inspection.

        Returns:
            Dictionary mapping tool names to their approval state
            ("allow_always" or "reject_always")
        """
        return self._tool_approvals.copy()


def _handle_enum_elicitation_response(
    response: RequestPermissionResponse, schema: dict[str, Any]
) -> types.ElicitResult | types.ErrorData:
    """Handle ACP response for enum elicitation."""
    from mcp import types

    if response.outcome.outcome == "selected":
        option_id = response.outcome.option_id

        if option_id == "cancel":
            return types.ElicitResult(action="cancel")

        # Extract enum value from option_id format: "enum_{index}_{value}"
        if option_id.startswith("enum_"):
            try:
                parts = option_id.split("_", 2)  # Split into ["enum", index, value]
                if len(parts) >= 3:  # noqa: PLR2004
                    enum_index = int(parts[1])
                    enum_values = schema.get("enum", [])
                    if 0 <= enum_index < len(enum_values):
                        selected_value = enum_values[enum_index]
                        return types.ElicitResult(action="accept", content=selected_value)
            except (ValueError, IndexError):
                pass

        # Fallback if parsing fails
        logger.warning("Failed to parse enum option_id", option_id=option_id)
        return types.ElicitResult(action="cancel")

    return types.ElicitResult(action="cancel")



"""Event handler configuration models for LLMling agent."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any, Literal

from pydantic import ConfigDict, Field
from schemez import Schema


if TYPE_CHECKING:
    from collections.abc import Sequence

    from pydantic_ai import RunContext

    from llmling_agent.agent.events import RichAgentStreamEvent
    from llmling_agent.common_types import IndividualEventHandler


class BaseEventHandlerConfig(Schema):
    """Base configuration for event handlers."""

    model_config = ConfigDict(
        frozen=True,
        arbitrary_types_allowed=True,
    )

    type: str = Field(init=False)
    """Event handler type discriminator."""

    enabled: bool = Field(default=True)
    """Whether this handler is enabled."""

    def get_handler(self) -> IndividualEventHandler:
        """Create and return the configured event handler.

        Returns:
            Configured event handler callable.

        Raises:
            NotImplementedError: Must be implemented by subclasses.
        """
        raise NotImplementedError


class BuiltinEventHandlerConfig(BaseEventHandlerConfig):
    """Configuration for built-in event handlers (simple, detailed)."""

    type: Literal["builtin"] = Field("builtin", init=False)
    """Builtin event handler."""

    handler: Literal["simple", "detailed"] = Field(
        default="simple",
        examples=["simple", "detailed"],
    )
    """Which builtin handler to use.

    - simple: Basic text and tool notifications
    - detailed: Comprehensive execution visibility
    """

    def get_handler(self) -> IndividualEventHandler:
        """Get the builtin event handler."""
        from llmling_agent.agent.builtin_handlers import (
            detailed_print_handler,
            simple_print_handler,
        )

        handlers = {
            "simple": simple_print_handler,
            "detailed": detailed_print_handler,
        }
        return handlers[self.handler]


class CallbackEventHandlerConfig(BaseEventHandlerConfig):
    """Configuration for custom callback event handlers via import path."""

    type: Literal["callback"] = Field("callback", init=False)
    """Callback event handler."""

    import_path: str = Field(
        examples=[
            "mymodule:my_handler",
            "mypackage.handlers:custom_event_handler",
        ],
    )
    """Import path to the handler function (module:function format)."""

    def get_handler(self) -> IndividualEventHandler:
        """Import and return the callback handler."""
        from llmling_agent.utils.importing import import_callable

        return import_callable(self.import_path)


class TTSEventHandlerConfig(BaseEventHandlerConfig):
    """Configuration for Text-to-Speech event handler with OpenAI streaming."""

    type: Literal["tts"] = Field("tts", init=False)
    """TTS event handler."""

    api_key: str | None = Field(
        default=None,
        examples=["sk-..."],
    )
    """OpenAI API key. If not provided, uses OPENAI_API_KEY env var."""

    model: Literal["tts-1", "tts-1-hd"] = Field(
        default="tts-1",
        examples=["tts-1", "tts-1-hd"],
    )
    """TTS model to use.

    - tts-1: Fast, optimized for real-time streaming
    - tts-1-hd: Higher quality, slightly higher latency
    """

    voice: Literal["alloy", "echo", "fable", "onyx", "nova", "shimmer"] = Field(
        default="alloy",
        examples=["alloy", "echo", "fable", "onyx", "nova", "shimmer"],
    )
    """Voice to use for synthesis."""

    speed: float = Field(
        default=1.0,
        ge=0.25,
        le=4.0,
        examples=[0.5, 1.0, 1.5, 2.0],
    )
    """Speed of speech (0.25 to 4.0, default 1.0)."""

    chunk_size: int = Field(
        default=1024,
        ge=256,
        examples=[512, 1024, 2048],
    )
    """Size of audio chunks to process (in bytes)."""

    sample_rate: int = Field(
        default=24000,
        examples=[16000, 22050, 24000, 44100],
    )
    """Audio sample rate in Hz (for PCM format)."""

    min_text_length: int = Field(
        default=20,
        ge=5,
        examples=[10, 20, 50],
    )
    """Minimum text length before synthesizing (in characters)."""

    def get_handler(self) -> IndividualEventHandler:  # noqa: PLR0915
        """Get the TTS event handler."""
        import asyncio
        import sys

        try:
            from openai import AsyncOpenAI
        except ImportError as e:
            msg = "openai package required for TTS. Install with: uv pip install llmling-agent[tts]"
            raise ImportError(msg) from e

        try:
            import sounddevice as sd  # type: ignore[import-untyped]
        except ImportError as e:
            msg = "sounddevice package required for TTS. Install extra 'tts'"
            raise ImportError(msg) from e

        from pydantic_ai import PartDeltaEvent, PartStartEvent
        from pydantic_ai.messages import TextPart, TextPartDelta

        from llmling_agent.agent.events import StreamCompleteEvent

        client = AsyncOpenAI(api_key=self.api_key)

        # Shared state for handler closure
        audio_queue: asyncio.Queue[bytes | None] = asyncio.Queue()
        playback_task: asyncio.Task[None] | None = None
        text_buffer = ""
        sentence_terminators = {".", "!", "?", "\n"}

        async def play_audio() -> None:
            """Async audio playback using sounddevice."""
            try:
                stream = sd.RawOutputStream(
                    samplerate=self.sample_rate,
                    channels=1,
                    dtype="int16",
                )
                stream.start()

                while True:
                    chunk = await audio_queue.get()
                    if chunk is None:
                        break
                    if chunk:
                        stream.write(chunk)

                stream.stop()
                stream.close()
            except Exception as e:  # noqa: BLE001
                print(f"\n❌ Audio playback error: {e}", file=sys.stderr)

        async def synthesize_text(text: str) -> None:
            """Synthesize text and queue audio chunks."""
            nonlocal playback_task

            if not text.strip():
                return

            # Start playback task if not running
            if playback_task is None or playback_task.done():
                playback_task = asyncio.create_task(play_audio())

            try:
                async with client.audio.speech.with_streaming_response.create(
                    model=self.model,
                    voice=self.voice,
                    input=text,
                    response_format="pcm",
                    speed=self.speed,
                ) as response:
                    async for chunk in response.iter_bytes(chunk_size=self.chunk_size):
                        await audio_queue.put(chunk)
            except Exception as e:  # noqa: BLE001
                print(f"\n❌ TTS error: {e}", file=sys.stderr)

        async def handler(ctx: RunContext, event: RichAgentStreamEvent[Any]) -> None:
            nonlocal text_buffer, playback_task

            match event:
                case (
                    PartStartEvent(part=TextPart(content=delta))
                    | PartDeltaEvent(delta=TextPartDelta(content_delta=delta))
                ):
                    text_buffer += delta

                    # Check for sentence boundaries
                    if any(term in text_buffer for term in sentence_terminators):
                        last_term = max(
                            (text_buffer.rfind(term) for term in sentence_terminators),
                            default=-1,
                        )

                        if last_term > 0 and last_term >= self.min_text_length:
                            sentence = text_buffer[: last_term + 1].strip()
                            text_buffer = text_buffer[last_term + 1 :]

                            if sentence:
                                await synthesize_text(sentence)

                case StreamCompleteEvent():
                    # Process remaining text
                    if text_buffer.strip():
                        await synthesize_text(text_buffer.strip())
                        text_buffer = ""

                    # Signal playback to stop
                    await audio_queue.put(None)
                    if playback_task and not playback_task.done():
                        await playback_task

        return handler


EventHandlerConfig = Annotated[
    BuiltinEventHandlerConfig | CallbackEventHandlerConfig | TTSEventHandlerConfig,
    Field(discriminator="type"),
]


def resolve_handler_configs(
    configs: Sequence[EventHandlerConfig] | None,
) -> list[IndividualEventHandler]:
    """Resolve event handler configs to actual handler callables.

    Args:
        configs: List of event handler configurations.

    Returns:
        List of resolved event handler callables.
    """
    if not configs:
        return []
    return [cfg.get_handler() for cfg in configs]



"""AG-UI remote agent implementation.

This module provides a MessageNode adapter that connects to remote AG-UI protocol servers,
enabling remote agent execution with streaming support.
"""

from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Self
from uuid import uuid4

from anyenv import MultiEventHandler
from anyenv.processes import hard_kill
import httpx
from pydantic import TypeAdapter

from llmling_agent.agent.agui_converters import (
    agui_to_native_event,
    convert_to_agui_content,
    extract_text_from_event,
    to_agui_input_content,
)
from llmling_agent.agent.builtin_handlers import resolve_event_handlers
from llmling_agent.agent.events import RunStartedEvent, StreamCompleteEvent
from llmling_agent.common_types import IndividualEventHandler
from llmling_agent.log import get_logger
from llmling_agent.messaging import ChatMessage, MessageHistory
from llmling_agent.messaging.messagenode import MessageNode
from llmling_agent.messaging.processing import prepare_prompts
from llmling_agent.talk.stats import MessageStats


if TYPE_CHECKING:
    from asyncio.subprocess import Process
    from collections.abc import AsyncIterator, Callable, Sequence
    from types import TracebackType

    from ag_ui.core import Event
    from evented.configs import EventConfig

    from llmling_agent.agent.events import RichAgentStreamEvent
    from llmling_agent.common_types import BuiltinEventHandlerType, PromptCompatible
    from llmling_agent.delegation import AgentPool
    from llmling_agent.messaging.context import NodeContext
    from llmling_agent_config.mcp_server import MCPServerConfig


logger = get_logger(__name__)


def get_client(headers: dict[str, str], timeout: float) -> httpx.AsyncClient:
    return httpx.AsyncClient(
        timeout=httpx.Timeout(timeout),
        headers={
            **headers,
            "Accept": "text/event-stream",
            "Content-Type": "application/json",
        },
    )


@dataclass
class AGUISessionState:
    """Track state for an active AG-UI session."""

    thread_id: str
    """Thread ID for this session."""
    run_id: str | None = None
    """Current run ID."""
    text_chunks: list[str] = field(default_factory=list)
    """Accumulated text chunks."""
    thought_chunks: list[str] = field(default_factory=list)
    """Accumulated thought chunks."""
    tool_calls: dict[str, dict[str, Any]] = field(default_factory=dict)
    """Active tool calls by ID."""
    is_complete: bool = False
    """Whether the current run is complete."""
    error: str | None = None
    """Error message if run failed."""

    def clear(self) -> None:
        """Clear session state."""
        self.text_chunks.clear()
        self.thought_chunks.clear()
        self.tool_calls.clear()
        self.is_complete = False
        self.error = None
        self.run_id = str(uuid4())


class AGUIAgent[TDeps = None](MessageNode[TDeps, str]):
    """MessageNode that wraps a remote AG-UI protocol server.

    Connects to AG-UI compatible endpoints via HTTP/SSE and provides the same
    interface as native agents, enabling composition with other nodes via
    connections, teams, etc.

    The agent manages:
    - HTTP client lifecycle (create on enter, close on exit)
    - AG-UI protocol communication via SSE streams
    - Event conversion to native llmling-agent events
    - Message accumulation and final response generation

    Supports both blocking `run()` and streaming `run_stream()` execution modes.

    Example:
        ```python
        # Connect to existing server
        async with AGUIAgent(
            endpoint="http://localhost:8000/agent/run",
            name="remote-agent"
        ) as agent:
            result = await agent.run("Hello, world!")
            async for event in agent.run_stream("Tell me a story"):
                print(event)

        # Start server automatically (useful for testing)
        async with AGUIAgent(
            endpoint="http://localhost:8000/agent/run",
            name="test-agent",
            startup_command="ag ui agent config.yml",
            startup_delay=2.0,
        ) as agent:
            result = await agent.run("Test prompt")
        ```
    """

    def __init__(
        self,
        endpoint: str,
        *,
        name: str = "agui-agent",
        description: str | None = None,
        display_name: str | None = None,
        timeout: float = 60.0,
        headers: dict[str, str] | None = None,
        startup_command: str | None = None,
        startup_delay: float = 2.0,
        mcp_servers: Sequence[str | MCPServerConfig] | None = None,
        agent_pool: AgentPool[Any] | None = None,
        enable_logging: bool = True,
        event_configs: Sequence[EventConfig] | None = None,
        event_handlers: Sequence[IndividualEventHandler | BuiltinEventHandlerType] | None = None,
    ) -> None:
        """Initialize AG-UI agent client.

        Args:
            endpoint: HTTP endpoint for the AG-UI agent
            name: Agent name for identification
            description: Agent description
            display_name: Human-readable display name
            timeout: Request timeout in seconds
            headers: Additional HTTP headers
            startup_command: Optional shell command to start server automatically.
                           Useful for testing - server lifecycle is managed by the agent.
                           Example: "ag ui agent config.yml"
            startup_delay: Seconds to wait after starting server before connecting (default: 2.0)
            mcp_servers: MCP servers to connect
            agent_pool: Agent pool for multi-agent coordination
            enable_logging: Whether to enable database logging
            event_configs: Event trigger configurations
            event_handlers: Sequence of event handlers to register
        """
        super().__init__(
            name=name,
            description=description,
            display_name=display_name,
            mcp_servers=mcp_servers,
            agent_pool=agent_pool,
            enable_logging=enable_logging,
            event_configs=event_configs,
        )
        self.endpoint = endpoint
        self.timeout = timeout
        self.headers = headers or {}
        # Startup command configuration
        self._startup_command = startup_command
        self._startup_delay = startup_delay
        self._startup_process: Process | None = None
        self._client: httpx.AsyncClient | None = None
        self._state: AGUISessionState | None = None
        self._message_count = 0
        self.conversation = MessageHistory()
        self._total_tokens = 0
        self._event_queue: asyncio.Queue[RichAgentStreamEvent[Any]] = asyncio.Queue()
        resolved_handlers = resolve_event_handlers(event_handlers)
        self.event_handler = MultiEventHandler[IndividualEventHandler](resolved_handlers)

    @property
    def context(self) -> NodeContext:
        """Get node context."""
        from llmling_agent.messaging.context import NodeContext
        from llmling_agent.models.manifest import AgentsManifest
        from llmling_agent_config.nodes import NodeConfig

        cfg = NodeConfig(name=self.name, description=self.description)
        defn = self.agent_pool.manifest if self.agent_pool else AgentsManifest()
        return NodeContext(node_name=self.name, pool=self.agent_pool, config=cfg, definition=defn)

    async def __aenter__(self) -> Self:
        """Enter async context - initialize client and base resources."""
        await super().__aenter__()
        self._client = get_client(self.headers, self.timeout)
        self._state = AGUISessionState(thread_id=self.conversation_id)
        if self._startup_command:  # Start server if startup command is provided
            await self._start_server()
        self.log.debug("AG-UI client initialized", endpoint=self.endpoint)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit async context - cleanup client and base resources."""
        if self._client:
            await self._client.aclose()
            self._client = None
        self._state = None
        if self._startup_process:  # Stop server if we started it
            await self._stop_server()
        self.log.debug("AG-UI client closed")
        await super().__aexit__(exc_type, exc_val, exc_tb)

    async def _start_server(self) -> None:
        """Start the AG-UI server subprocess."""
        if not self._startup_command:
            return

        self.log.info("Starting AG-UI server", command=self._startup_command)
        self._startup_process = await asyncio.create_subprocess_shell(
            self._startup_command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            start_new_session=True,  # Create new process group
        )
        self.log.debug("Waiting for server startup", delay=self._startup_delay)
        await asyncio.sleep(self._startup_delay)
        # Check if process is still running
        if self._startup_process.returncode is not None:
            stderr = ""
            if self._startup_process.stderr:
                stderr = (await self._startup_process.stderr.read()).decode()
            msg = f"Startup process exited with code {self._startup_process.returncode}: {stderr}"
            raise RuntimeError(msg)

        self.log.info("AG-UI server started")

    async def _stop_server(self) -> None:
        """Stop the AG-UI server subprocess."""
        if not self._startup_process:
            return

        self.log.info("Stopping AG-UI server")
        try:
            await hard_kill(self._startup_process)  # Use cross-platform hard kill helper
        except Exception:  # Log but don't fail if kill has issues
            self.log.exception("Error during process termination")
        finally:
            self._startup_process = None
            self.log.info("AG-UI server stopped")

    async def run(
        self,
        *prompts: PromptCompatible,
        message_id: str | None = None,
        message_history: MessageHistory | None = None,
        **kwargs: Any,
    ) -> ChatMessage[str]:
        """Execute prompt against AG-UI agent.

        Sends the prompt to the AG-UI server and waits for completion.
        Events are collected via run_stream and event handlers are called.
        The final text content is returned as a ChatMessage.

        Args:
            prompts: Prompts to send (will be joined with spaces)
            message_id: Optional message id for the returned message
            message_history: Optional MessageHistory to use instead of agent's own
            **kwargs: Additional arguments (ignored for compatibility)

        Returns:
            ChatMessage containing the agent's aggregated text response
        """
        final_message: ChatMessage[str] | None = None
        async for event in self.run_stream(
            *prompts, message_id=message_id, message_history=message_history
        ):
            if isinstance(event, StreamCompleteEvent):
                final_message = event.message

        if final_message is None:
            raise RuntimeError("No final message received from stream")
        return final_message

    async def run_stream(
        self,
        *prompts: PromptCompatible,
        message_id: str | None = None,
        message_history: MessageHistory | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[RichAgentStreamEvent[str]]:
        """Execute prompt with streaming events.

        Args:
            prompts: Prompts to send
            message_id: Optional message ID
            message_history: Optional MessageHistory to use instead of agent's own
            **kwargs: Additional arguments (ignored for compatibility)

        Yields:
            Native streaming events converted from AG-UI protocol
        """
        from ag_ui.core import RunAgentInput, UserMessage

        if not self._client or not self._state:
            msg = "Agent not initialized - use async context manager"
            raise RuntimeError(msg)

        conversation = message_history if message_history is not None else self.conversation
        user_msg, _processed_prompts, _original_message = await prepare_prompts(*prompts)
        self._state.clear()  # Reset state
        run_started = RunStartedEvent(
            thread_id=self._state.thread_id,
            run_id=self._state.run_id or str(uuid4()),
            agent_name=self.name,
        )
        for handler in self.event_handler._wrapped_handlers:
            await handler(None, run_started)
        yield run_started

        # Get pending parts from conversation and convert them
        pending_parts = conversation.get_pending_parts()
        pending_content = to_agui_input_content(pending_parts)

        # Convert prompts to AGUI content (always returns list)
        content = await convert_to_agui_content(prompts)

        # Combine pending parts with new content
        final_content = [*pending_content, *content]

        user_message = UserMessage(id=str(uuid4()), content=final_content)
        request_data = RunAgentInput(
            thread_id=self._state.thread_id,
            run_id=self._state.run_id,
            state={},
            messages=[user_message],
            tools=[],
            context=[],
            forwarded_props={},
        )

        self.log.debug("Sending prompt to AG-UI agent")
        data = request_data.model_dump(by_alias=True)
        try:  # Send request and stream events
            async with self._client.stream("POST", self.endpoint, json=data) as response:
                response.raise_for_status()
                async for event in self._parse_sse_stream(response):
                    # Track text chunks
                    if text := extract_text_from_event(event):
                        self._state.text_chunks.append(text)

                    # Convert to native event and distribute to handlers
                    if native_event := agui_to_native_event(event):
                        # Check for queued custom events first
                        while not self._event_queue.empty():
                            try:
                                custom_event = self._event_queue.get_nowait()
                                for handler in self.event_handler._wrapped_handlers:
                                    await handler(None, custom_event)
                                yield custom_event
                            except asyncio.QueueEmpty:
                                break
                        # Distribute to handlers
                        for handler in self.event_handler._wrapped_handlers:
                            await handler(None, native_event)
                        yield native_event

        except httpx.HTTPError as e:
            self._state.error = str(e)
            self.log.exception("HTTP error during AG-UI run")
            raise
        finally:
            self._state.is_complete = True
            # Emit final message
            final_message = ChatMessage[str](
                content="".join(self._state.text_chunks),
                role="assistant",
                name=self.name,
                message_id=message_id or str(uuid4()),
                conversation_id=self.conversation_id,
            )
            complete_event = StreamCompleteEvent(message=final_message)
            for handler in self.event_handler._wrapped_handlers:
                await handler(None, complete_event)
            yield complete_event
            # Record to conversation history
            conversation.add_chat_messages([user_msg, final_message])

    async def _parse_sse_stream(self, response: httpx.Response) -> AsyncIterator[Event]:
        """Parse Server-Sent Events stream.

        Args:
            response: HTTP response with SSE stream

        Yields:
            Parsed AG-UI events
        """
        from ag_ui.core import Event

        event_adapter: TypeAdapter[Event] = TypeAdapter(Event)
        buffer = ""
        async for chunk in response.aiter_text():
            buffer += chunk
            # Process complete SSE events
            while "\n\n" in buffer:
                event_text, buffer = buffer.split("\n\n", 1)
                # Parse SSE format: "data: {json}\n"
                for line in event_text.split("\n"):
                    if line.startswith("data: "):
                        json_str = line[6:]  # Remove "data: " prefix
                        try:
                            event = event_adapter.validate_json(json_str)
                            yield event
                        except (ValueError, TypeError) as e:
                            self.log.warning(
                                "Failed to parse AG-UI event",
                                json=json_str[:100],
                                error=str(e),
                            )

    async def run_iter(
        self,
        *prompt_groups: Sequence[PromptCompatible],
        message_id: str | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[ChatMessage[str]]:
        """Execute multiple prompt groups sequentially.

        Args:
            prompt_groups: Groups of prompts to execute
            message_id: Optional message ID base
            **kwargs: Additional arguments (ignored for compatibility)

        Yields:
            ChatMessage for each completed prompt group
        """
        for i, prompts in enumerate(prompt_groups):
            mid = f"{message_id or 'msg'}_{i}" if message_id else None
            yield await self.run(*prompts, message_id=mid)

    def to_tool(self, description: str | None = None) -> Callable[[str], Any]:
        """Convert agent to a callable tool.

        Args:
            description: Tool description

        Returns:
            Async function that can be used as a tool
        """

        async def wrapped(prompt: str) -> str:
            """Execute AG-UI agent with given prompt."""
            result = await self.run(prompt)
            return result.content

        wrapped.__name__ = self.name
        wrapped.__doc__ = description or f"Call {self.name} AG-UI agent"
        return wrapped

    @property
    def model_name(self) -> str | None:
        """Get model name (AG-UI doesn't expose this)."""
        return None

    async def get_stats(self) -> MessageStats:
        """Get message statistics for this node."""
        return MessageStats()


async def main() -> None:
    """Example usage."""
    async with AGUIAgent(endpoint="http://localhost:8000/agent/run", name="test-agent") as agent:
        result = await agent.run("What is 2+2?")
        print(f"Result: {result.content}")
        print("\nStreaming:")
        async for event in agent.run_stream("Tell me a short joke"):
            print(f"Event: {event}")


if __name__ == "__main__":
    asyncio.run(main())
